"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import yaml
import json
import dill
import copy
from collections import OrderedDict
from .utils import log_message, b64encode, b64decode, get_logger
from .pipeline_loader import PipelineLoader

log = get_logger(__name__)

# Configure dill to 'recurse' dependencies when dumping objects
dill.settings['recurse'] = True


class _FunctionStep:

    def __init__(self, name, function_name, function_type, serialized_code):
        self.name = name
        self.function_name = function_name
        self.function_type = function_type
        self.serialized_code = serialized_code

    def run(self, pipeline, data, code_deserializer):
        if self.function_type == 'inline':
            fn = code_deserializer(self.serialized_code)
            log_message('> %s ' % self.function_name, log)
            result = fn(pipeline, data)
            if result is not None:
                data = result
            return data

    def to_camel(self, camel='1.0.0'):
        return {'name': self.name, 'function': {'name': self.function_name, 'code': b64encode(self.serialized_code), 'type': self.function_type}}

    @staticmethod
    def load(doc):
        name = doc.get('name')
        f = doc.get('function', {})
        function_name = f.get('name')
        function_type = f.get('type')
        code = f.get('code')

        return _FunctionStep(name, function_name, function_type, b64decode(code))


class _TransformStep:

    def __init__(self, name, module_name, class_name):
        self.class_name = class_name
        self.module_name = module_name
        self.full_name = '.'.join([module_name, class_name])
        self.name = name or self.full_name
        self._validate_transform(module_name, class_name)

    def run(self, pipeline, data, code_deserializer):
        t = self._get_instance(self.module_name, self.class_name)
        log_message('> %s ' % self.name, log)
        result = t.transform(pipeline, data)
        if result is not None:
            data = result
        return data

    def to_camel(self, camel='1.0.0'):
        return {'name': self.name, 'transform': {'class_name': self.class_name, 'module_name': self.module_name}}

    @staticmethod
    def load(doc):
        name = doc.get('name')
        t = doc.get('transform', {})
        class_name = t.get('class_name')
        module_name = t.get('module_name')

        return _TransformStep(name, module_name, class_name)

    @staticmethod
    def _get_instance(module_name, class_name):
        try:
            # Attempt to import module
            return getattr(__import__(module_name, globals(), locals(), [class_name], 0), class_name)()
        except ValueError:
            # Attempt to locate class in globals
            return globals()[class_name]()

    @staticmethod
    def _validate_transform(module_name, class_name):
        full_name = '.'.join([module_name, class_name])

        # Transform validation
        # 1. Make sure we can instantiate the step
        t = _TransformStep._get_instance(module_name, class_name)

        # 2. Make sure the class has a 'transform' attribute that is callable
        try:
            t_func = getattr(t, 'transform')
            if not callable(t_func):
                raise ValueError('"transform" attribute of class ' + full_name + ' is not callable')
        except AttributeError:
            raise AttributeError('Provided class ' + full_name + ' does not have a "transform" method')


class Pipeline:

    """
    Provides a pipeline abstraction used to transform or enrich data; steps either Python functions
    or classes with a transform method; a pipeline step accepts a dataframe as an argument and is expected to transform
    or enrich the dataframe in sequential order.

    To create a pipeline to remove an unused column:

        >>> from cortex import Cortex
        >>> train_ds = cortex.dataset(' <path to your data set> ')
        >>> pipeline = train_ds.pipeline(' <pipeline name> ', clear_cache=True)
        >>> pipeline.reset()
        >>> def drop_unused(pipeline, df):
        ...     df.drop(columns=['Id'], axis=1, inplace=True) # this pipeline will drop the field ID
        ...
        >>> pipeline.add_step(drop_unused)

    """

    def __init__(self, name: str, depends=None, loader=PipelineLoader.default_loader()):
        self._name = name
        self._dependencies = OrderedDict({})
        self._steps = []
        self._context = {}
        self._loader = loader

        if depends is not None and len(depends) > 0:
            for d in depends:
                self.add_dependency(loader.get_pipeline(d))

        loader.add_pipeline(name, self)

    @property
    def name(self):
        """
        Gets the name of the pipeline.
        """
        return self._name

    @property
    def steps(self):
        """
        Gets a list of steps in the pipeline.
        """
        return self._steps

    @property
    def dependencies(self):
        """
        Gets a list of dependencies for the pipeline.
        """
        return self._dependencies.values()

    def get_context(self, key: str, default_value=None, deserializer=dill.loads):
        """
        Gets the context for the pipeline.

        :param key: The key for the context.
        :param default_value: (Optional) A default value to return if the context for the given key is not found
        :return: The context for the given key, if it exists. If it does not exist, the default value or `none`.
        """
        if deserializer is None:
            def _identity(o): return o
            deserializer = _identity

        val = self._context.get(key)
        if val:
            return deserializer(val)
        else:
            depends = self.dependencies
            if len(depends) > 0:
                for dep in depends:
                    if self._name == dep._name:
                        raise Exception('Circular dependency detected in pipeline dependency graph')
                    val = dep.get_context(key, deserializer=None)
                    if val:
                        return deserializer(val)

        return default_value

    def set_context(self, key: str, obj, serializer=dill.dumps):
        """
        Sets the context for the pipeline.

        :param key: The key for the context to set.
        :param obj: The value for the context to set.
        """
        if serializer is None:
            def _identity(o): return o
            serializer = _identity

        self._context[key] = serializer(obj)

    def add_step(self, step, name=None, serializer=dill.dumps):
        """
        Adds a transformation function to the pipeline.

        :param fn: The function to add.
        :param name: An optional name for the function to add.
        """
        if isinstance(step, type):
            return self._add_transform_step(step, name)
        return self._add_inline_step(step, name, serializer)

    def _add_inline_step(self, fn, name=None, serializer=dill.dumps):
        fn_name = fn.__name__
        if name is None:
            name = fn_name

        code = serializer(fn)
        self._steps.append(_FunctionStep(name, fn_name, 'inline', code))

        return self

    def _add_transform_step(self, kls, name=None):
        class_name = kls.__name__
        module_name = kls.__module__
        self._steps.append(_TransformStep(name, module_name, class_name))

        return self

    def get_step(self, name, deserializer=dill.loads):
        """
        Gets a pipeline step by name.

        :param name: Name of the function to get.

        :return: The code for the step.
        """
        for step in self._steps:
            if step.name == name:
                if isinstance(step, _FunctionStep):
                    return deserializer(step.serialized_code)
                elif isinstance(step, _TransformStep):
                    return step._get_instance(step.module_name, step.class_name)
                else:
                    raise ValueError('Invalid step: must be either a function or a transform class')
        return None

    def remove_step(self, name):
        """
        Removes a step by name.

        :param name: Name of the function to remove.
        """
        new_steps = []
        for step in self._steps:
            if step.name != name:
                new_steps.append(step)

        self._steps = new_steps

        return self

    def add_dependency(self, pipeline):
        """
        Adds a dependency on another pipeline.

        :param pipeline: the pipeline on which this pipeline depends.
        """
        self._dependencies[pipeline.name] = pipeline
        return self

    def reset(self, reset_deps=False, reset_context=False):
        """
        Removes all steps from the pipeline, but leaves context and dependencies in place. Context and
        dependencies can also be reset using the optional flags for each.

        :param reset_deps: If `True`, removes all dependencies.  Default: `False`.
        :param reset_context: If `True`, resets context. Default: `False`
        :return: The pipeline instance.
        """
        self._steps = []
        if reset_deps:
            self._dependencies = OrderedDict({})
        if reset_context:
            self._context = {}

        return self

    def from_pipeline(self, pipeline):
        """
        Copies a pipeline from the given pipeline.

        :param pipeline: The pipeline from which to copy this pipeline.
        """
        self._steps = copy.deepcopy(pipeline.steps)
        self._dependencies = copy.deepcopy(pipeline._dependencies)
        self._context = copy.deepcopy(pipeline._context)
        return self

    def _run_dependencies(self, data):
        depends = self.dependencies
        if len(depends) > 0:
            for dep in depends:
                if self._name == dep._name:
                    raise Exception('Circular dependency detected in pipeline dependency graph')

                data = dep.run(data)

        return data

    def run(self, data, code_deserializer=dill.loads):
        """
        Runs the steps on the pipeline.

        :param data: Data to pass to dependencies.
        :return: The dataframe as transformed by the dependencies and steps.
        """
        # run dependencies
        df = self._run_dependencies(data)

        log_message('running pipeline [%s]:' % self._name, log)

        for step in self._steps:
            df = step.run(self, df, code_deserializer)

        return df

    def to_camel(self, camel='1.0.0'):
        pipeline = {
            'name': self._name,
            'steps': [step.to_camel(camel) for step in self._steps],
            'dependencies': [dep.name for dep in self.dependencies],
            'context': {k: b64encode(v) for k, v in self._context.items()}
        }
        return pipeline

    def dumps(self, stream=None, notebook=False, camel='1.0.0', yaml_format=False):
        """
        Dumps the steps, dependencies, and context for the pipeline.

        :param stream: Optional stream on which to dump the pipeline.
        :param notebook: Optional notebook where the dump should be displayed.
        """
        if stream is None and notebook:
            stream = sys.stdout

        doc = self.to_camel(camel)
        print(doc)

        if yaml_format:
            s = yaml.dump(doc, stream=stream, indent=2)
            if stream is None:
                return s
        else:
            if stream is None:
                return json.dumps(doc)
            else:
                json.dump(doc, stream)

    @staticmethod
    def load(doc, loader=PipelineLoader.default_loader()):
        p = Pipeline(name=doc.get('name'), loader=loader)
        steps = doc.get('steps', [])
        context = doc.get('context', {})
        deps = doc.get('dependencies', [])

        for step in steps:
            if step.get('function'):
                p._steps.append(_FunctionStep.load(step))
            elif step.get('transform'):
                p._steps.append(_TransformStep.load(step))

        for key in context.keys():
            p._context[key] = b64decode(context[key])

        for d in deps:
            p._dependencies[d] = loader.get_pipeline(d)

        return p



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class CortexException(Exception):
    """
    Base exception type.
    """
    pass


class AuthenticationException(CortexException):
    """
    Cortex authentication exception.
    """
    pass


class BadTokenException(CortexException):
    """
    Cortex token invalid exception.
    """
    pass


class ConfigurationException(CortexException):
    """
    Cortex configuration exception.
    """
    pass


class APIException(CortexException):
    """
    Cortex API exception.
    """
    pass



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import cuid
import io
import os
import dill
import tempfile
from .timer import Timer
from .exceptions import ConfigurationException


class Run:
    """
    Captures the elapsed time of a run of an experiment and saves the metrics and parameters for the run.
    """
    def __init__(self, experiment):
        self._id = cuid.slug()
        self._experiment = experiment
        self._timer = None
        self._start = None
        self._end = None
        self._interval = None
        self._params = {}
        self._metrics = {}
        self._artifacts = {}
        self._meta = {}

    def start(self):
        """
        Starts the timer.
        """
        if self._timer is not None:
            raise ValueError('Attempt to start a Run that is already started')

        self._timer = Timer()
        self._timer.start()
        self._start = self._timer.start_time
        self._experiment.save_run(self)

    def stop(self):
        """
        Stops the time and captures data about the run.
        """
        self._timer.stop()
        self._end = self._timer.end_time
        self._interval = self._timer.interval
        self._experiment.save_run(self)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    @property
    def id(self):
        """
        The id for the run.
        """
        return self._id

    @property
    def start_time(self):
        """
        The start time for the run.
        """
        if self._start:
            return int(self._start)
        return None

    @property
    def end_time(self):
        """
        The end time for the run.
        """
        if self._end:
            return int(self._end)
        return None

    @property
    def took(self):
        """
        The interval between the start and end of the run.
        """
        return self._interval

    @property
    def params(self):
        """
        Params of the run.
        """
        return self._params

    def get_param(self, param):
        """
        Gets a particular parameter of the run.

        :param param: Parameter to get.
        """
        return self._params.get(param)

    @property
    def metrics(self):
        """
        The metrics for the run.
        """
        return self._metrics

    def get_metric(self, metric):
        """
        Gets a particular metric for the run.

        :param metric: metric to get
        """
        return self._metrics.get(metric)

    @property
    def meta(self):
        """
        Metadata for the run.
        """
        return self._meta

    def get_meta(self, meta):
        """
        Gets metadata for the run.
        """
        return self.meta.get(meta)

    @property
    def artifacts(self):
        """
        The artifacts stored in this run.

        :return: The artifact object.
        """
        return self._artifacts

    def get_artifact(self, name):
        """
        Gets an artifact object by name.

        :param name: The key of the artifact.
        :return: The artifact object.
        """
        return self.artifacts.get(name)

    def to_json(self):
        """
        A json representation of the run.
        """
        doc = {'id': self.id, 'startTime': self.start_time, 'endTime': self.end_time, 'took': self.took,
               'params': self.params, 'metrics': self.metrics, 'meta': self.meta,
               'artifacts': [name for name in self.artifacts.keys()]}
        return doc

    @staticmethod
    def from_json(json, experiment):
        """
        Creates a run from a json representation.

        :param json: The json representation of the run.
        :param experiment: The experiment to associate with the run.
        :return: A run that has the values in the given json object with
        the given experiment.
        """
        run = Run(experiment)
        run._id = json['id']
        run._start = json.get('startTime', json.get('start'))
        run._end = json.get('endTime', json.get('end'))
        run._interval = json.get('took')
        run._params = json.get('params', {})
        run._metrics = json.get('metrics', {})
        run._meta = json.get('meta', {})

        artifacts = json.get('artifacts', [])
        for name in artifacts:
            run._artifacts[name] = experiment.load_artifact(run, name)

        return run

    def log_param(self, name: str, param):
        """
        Logs for a given parameter.
        """
        if hasattr(param, 'tolist'):
            _val = param.tolist()
        else:
            _val = self._to_scaler(param)

        self._params[name] = _val

    def log_params(self, params):
        """
        Logs a group of parameters.
        """
        for k, v in params.items():
            self.log_param(k, v)

    def log_metric(self, name: str, metric):
        """
        Logs a given metric.
        """
        if hasattr(metric, 'tolist'):
            _val = metric.tolist()
        else:
            _val = self._to_scaler(metric)

        self._metrics[name] = _val

    def set_meta(self, name: str, val):
        """
        Sets metadata for the run.
        """
        if hasattr(val, 'tolist'):
            _val = val.tolist()
        else:
            _val = self._to_scaler(val)

        self._meta[name] = _val

    def log_artifact(self, name: str, artifact):
        """
        Logs a given artifact.
        """
        self._artifacts[name] = artifact

    def log_artifact_ref(self, name: str, ref):
        """
        Logs a given artifact reference.
        """
        self._artifacts[name] = {'ref': ref}

    @staticmethod
    def _is_numpy_dtype(x):
        return hasattr(x, 'dtype')

    @staticmethod
    def _to_scaler(x):
        if hasattr(x, 'dtype'):
            import numpy as np
            return np.asscalar(x)
        return x


class RemoteRun(Run):
    """
    A run that is executed remotely, through a client.
    """

    def __init__(self, experiment, experiment_client):
        super().__init__(experiment)
        self._client = experiment_client

    @staticmethod
    def create(experiment, experiment_client):
        """
        Creates a remote run.

        :param experiment: The experiment to associate with this run.
        :param experiment_client: The client for the run.
        :return: A run.
        """
        r = experiment_client.create_run(experiment.name)
        run = RemoteRun(experiment, experiment_client)
        run._id = r['runId']

        return run

    @staticmethod
    def get(experiment, run_id, experiment_client):
        """
        Gets a run.

        :param experiment: The parent experiment of the run.
        :param run_id: The identifier for the run.
        :param experiment_client: The client for the run.
        :return: A run.
        """
        r = experiment_client.get_run(experiment.name, run_id)
        return RemoteRun.from_json(r, experiment)

    @staticmethod
    def from_json(json, experiment):
        """
        Builds a run from the given json.

        :param jsom: json that specifies the run; acceptable values are runId,
        startTime, endTime, took, a list of params, metrics, metadata, and artifacts
        :param experiment: the parent experiment of the run
        :return: a run
        """
        run = RemoteRun(experiment, experiment._client)
        run._id = json['runId']
        run._start = json.get('startTime', json.get('start'))
        run._end = json.get('endTime', json.get('end'))
        run._interval = json.get('took')
        run._params = json.get('params', {})
        run._metrics = json.get('metrics', {})
        run._meta = json.get('meta', {})
        run._artifacts = json.get('artifacts', {})

        return run

    def log_param(self, name: str, param):
        """
        Updates the params for the run.
        """
        super().log_param(name, param)
        self._client.update_param(self._experiment.name, self.id, name, param)

    def log_metric(self, name: str, metric):
        """
        Updates the metrics for the run.
        """
        super().log_metric(name, metric)
        self._client.update_metric(self._experiment.name, self.id, name, metric)

    def set_meta(self, name: str, val):
        """
        Sets the metadata for the run.
        """
        super().set_meta(name, val)
        self._client.update_meta(self._experiment.name, self.id, name, val)

    def log_artifact(self, name: str, artifact):
        """
        Updates the artifacts for the run.
        """
        super().log_artifact(name, artifact)
        if hasattr(artifact, 'ref'):
            with open(artifact['ref'], 'rb') as stream:
                self.log_artifact_stream(name, stream)
        else:
            stream = io.BytesIO()
            dill.dump(artifact, stream)
            stream.seek(0)
            self.log_artifact_stream(name, stream)

    def log_artifact_file(self, name: str, file_path):
        """
        Logs the artifact to the file given in the filepath.
        """
        super().log_artifact(name, file_path)
        with open(file_path, 'rb') as f:
            self.log_artifact_stream(name, f)

    def log_artifact_stream(self, name: str, stream):
        """
        Updates the artifact with the given stream.
        """
        self._client.update_artifact(self._experiment.name, self.id, name, stream)

    def log_keras_model(self, model, artifact_name='model'):
        """
        Logs a keras model as an artifact.
        """
        with tempfile.NamedTemporaryFile(mode='w+b') as temp:
            model.save(filepath=temp.name)
            self.log_artifact_file(artifact_name, temp.name)

    def get_artifact(self, name: str, deserializer=dill.loads):
        """
        Gets an artifact with the given name.  Deserializes the artifact stream using dill by default.  Deserialization
        can be disabled entirely or the deserializer function can be overridden.
        """
        artifact_bytes = self._client.get_artifact(self._experiment.name, self.id, name)
        if deserializer:
            return deserializer(artifact_bytes)
        return artifact_bytes

    def get_keras_model(self, artifact_name='model'):
        """
        Gets the keras model.
        """
        try:
            from keras.models import load_model
        except ImportError:
            raise ConfigurationException('Keras needs to be installed in order to use get_keras_model')

        f = tempfile.NamedTemporaryFile(delete=False)
        f.write(self.get_artifact(artifact_name, deserializer=lambda x: x))
        model_file = f.name
        f.close()
        try:
            model = load_model(model_file)
        finally:
            os.remove(model_file)

        return model



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import jwt
import base64
import hashlib
import logging
from requests.exceptions import HTTPError
from pathlib import Path
from datetime import datetime
from collections import namedtuple
from .exceptions import BadTokenException


def md5sum(file_name, blocksize=65536):
    """
    Computes md5sum on a fileself.

    :param file_name: The name of the file on which to compute the md5sum.
    :return: Hexdigest of md5sum for file.
    """
    md5 = hashlib.md5()
    with open(file_name, "rb") as f:
        for block in iter(lambda: f.read(blocksize), b""):
            md5.update(block)
    return md5.hexdigest()


def is_notebook() -> bool:
    """
    Checks if the shell is in a notebook.

    :return: `True` if the shell is for a notebook, `False` otherwise
    """
    try:
        from IPython import get_ipython
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or console
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except (NameError, ImportError):
        return False


def log_message(msg: str, log: logging.Logger, level=logging.INFO, *args, **kwargs):
    """
    Logs a message.

    :param msg: Message to log
    :param log: logger where message should be logged
    :param level: optional log level, defaults to INFO
    :param args: a tuple of arguments passed to the logger
    :param kwargs: a dictionary of keyword arguments passed to the logger
    """
    if is_notebook():
        print(msg)
        log.debug(msg, *args, **kwargs)
    else:
        log.log(level, msg, *args, **kwargs)


def b64encode(b: bytes)->str:
    """
    Returns a string from an iterable collection of bytes.
    """
    encoded = base64.b64encode(b)
    return encoded.decode('utf-8')


def b64decode(s: str)->bytes:
    """
    Returns an iterable collection of bytes representing a base-64 encoding of a given string.
    """
    return base64.decodebytes(s.encode('utf-8'))


def named_dict(obj):
    """
    Returns a named tuple for the given object if it's a dictionary or list;
    otherwise it returns the object itself.
    """
    if isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = named_dict(value)
        return namedtuple('NamedDict', obj.keys())(**obj)
    elif isinstance(obj, list):
        return [named_dict(item) for item in obj]
    else:
        return obj


def decode_JWT(*args, verify):
    """
    thin wrapper around jwt.decode. This function exists for better error handling of the
    jwt exceptions.
    """
    invalidTokenMsg = 'Your Cortex Token is invalid. For more information, go to Cortex Docs > Cortex Tools > Access'
    expiredTokenMsg = 'Your Cortex Token has expired. For more information, go to Cortex Docs > Cortex Tools > Access'
    try:
        decodedJWT = jwt.decode(*args, verify=verify)
        # there are places in the sdk where we try to decode 'any ol token' before sending the token to kong to get verified
        # therefore, here we have some reasonable checks to make sure that this is a cortex token by checking the JWT keys exist
        if not decodedJWT.get('tenant') or not decodedJWT.get('sub') or not decodedJWT.get('exp'):
            raise BadTokenException(invalidTokenMsg)
        if datetime.today().timestamp() > decodedJWT['exp']:
            raise jwt.ExpiredSignatureError
        return decodedJWT
    except jwt.ExpiredSignatureError:
        raise BadTokenException(expiredTokenMsg)
    except jwt.exceptions.InvalidTokenError:
        raise BadTokenException(invalidTokenMsg)


def get_cortex_profile(profile_name=None):
    """
    Gets the current cortex profile or the profile that matches the optionaly given name.
    """
    cortex_config_path = Path.home() / '.cortex/config'

    if cortex_config_path.exists():
        with cortex_config_path.open() as f:
            cortex_config = json.load(f)

        if profile_name is None:
            profile_name = cortex_config.get('currentProfile')

        return cortex_config.get('profiles', {}).get(profile_name, {})
    return {}


def get_logger(name):
    """
    Gets a logger with the given name.
    """
    log = logging.getLogger(name)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s/%(module)s: %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    log.addHandler(handler)
    log.setLevel(logging.INFO)
    return log


def json_str(val):
    """
    Get the string representation of a json object.
    """
    try:
        return json.dumps(val)
    except TypeError:
        return str(val)


def base64decode_jsonstring(base64encoded_jsonstring:str):
    """
    Loads a json from a base64 encoded json string.
    :param base64encoded_jsonstring:
    :return:
    """
    return json.loads(base64.urlsafe_b64decode(base64encoded_jsonstring).decode('utf-8'))


def raise_for_status_with_detail(resp):
    """
    wrap raise_for_status and attempt give detailed reason for api failure
    re-raise HTTPError for normal flow
    :param resp: python request resp
    :return:
    """
    try:
        resp.raise_for_status()
    except HTTPError as http_exception:
        try:
            log_message(msg=resp.json(), log=get_logger('http_status'), level=logging.ERROR)
        except Exception as e:
            pass # resp.json() failed
        finally:
            raise http_exception




"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .message import Message
from .utils import get_logger
from .action import ActionClient
from .serviceconnector import ServiceConnector
from .camel import CamelResource
from .utils import raise_for_status_with_detail

log = get_logger(__name__)


class Skill(CamelResource):
    """
     Computational components of an agent; executes an
     atomic unit of work and can be triggered by one or more inputs to produce
     one or more outputs.
    """

    def __init__(self, skill, connector: ServiceConnector):
        super().__init__(skill, True)
        self._connector = connector

    def _merge_properties(self, msg_props):
        if self.properties:
            for prop in self.properties:
                name = prop.get('name')
                value = prop.get('value', prop.get('defaultValue'))

                if name in msg_props or not value:
                    continue

                msg_props[name] = value

        return msg_props

    def invoke(self, input_name: str, message: Message, timeout=30):
        """
        Invokes the skill with a given message.

        :param input_name: Identifier for an input to this skill.
        :param message: The message used for the invocation.
        """
        for input in self.inputs:
            if input['name'] == input_name:
                routing = input['routing']
                action_name = routing.get('all', {}).get('action')
                if action_name:
                    action_client = ActionClient(self._connector.url, self._connector.version, self._connector.token)

                    # Prepare message properties
                    if not message.properties:
                        message.properties = {}

                    # Add Skill properties that are set to the message
                    message.properties = self._merge_properties(message.properties)

                    return Message(action_client.invoke_action(action_name, message.to_params()))
                else:
                    raise Exception('Skill invoke only works with "all" routing')
        else:
            raise Exception('No inputs defined')

    @staticmethod
    def get_skill(name: str, connector: ServiceConnector):
        """
        Fetches a skill to work with.

        :param name: The name of the skill to retrieve.
        :param connector: connector to use
        :return: A skill object.
        """
        uri = 'catalog/skills/{name}'.format(name=name)
        log.debug('Getting skill using URI: %s' % uri)
        r = connector.request('GET', uri)
        raise_for_status_with_detail(r)

        return Skill(r.json(), connector)



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import unittest

from mocket.mockhttp import Entry
from mocket import mocketize

from cortex import Cortex, Message
from cortex.agent import Agent, ServiceActivation
from .fixtures import john_doe_token, build_mock_url, mock_api_endpoint


class TestAgent(unittest.TestCase):

    def setUp(self):
        self.cortex = Cortex.client(api_endpoint=mock_api_endpoint(), api_version=3, token=john_doe_token())

    def register_entry(self, verb, url, body):
        print('Registering mock for', verb, url)
        Entry.single_register(verb,
                              url,
                              status=200,
                              body=json.dumps(body))

    @mocketize
    def test_get_agent(self):
        agent_name = 'unittest/agent1'
        uri = Agent.URIS['get_agent'].format(name=agent_name)
        returns = {"name": agent_name}
        self.register_entry(Entry.GET, build_mock_url(uri), returns)

        agent = self.cortex.agent(agent_name)
        self.assertEqual(agent.name, agent_name)

    @mocketize
    def test_invoke_sync(self):
        agent_name = 'unittest/agent1'
        service_name = 'input1'
        activation_id = '123456789'
        activation = {'activationId': activation_id, 'status': 'COMPLETE', 'response': {'text': 'It works'}}
        activation_msg = {'activation': activation}

        agent_uri = Agent.URIS['get_agent'].format(name=agent_name)
        self.register_entry(Entry.GET, build_mock_url(agent_uri), {"name": agent_name})

        invoke_uri = Agent.URIS['invoke_service'].format(agent_name=agent_name, service_name=service_name)
        self.register_entry(Entry.POST, build_mock_url(invoke_uri), activation)

        activation_uri = ServiceActivation.URIS['activation'].format(activation_id=activation_id)
        self.register_entry(Entry.GET, build_mock_url(activation_uri), activation_msg)

        agent = self.cortex.agent(agent_name)
        self.assertEqual(agent.name, agent_name)

        msg = agent.invoke_service(service_name, self.cortex.message({'text': 'Testing...'}))
        self.assertEqual(msg.payload.get('text'), 'It works')

    @mocketize
    def test_invoke_async(self):
        agent_name = 'unittest/agent1'
        service_name = 'input1'
        activation_id = '123456789'
        activation = {'activationId': activation_id, 'status': 'COMPLETE', 'response': {'text': 'It works'}}
        activation_msg = {'activation': activation}

        agent_uri = Agent.URIS['get_agent'].format(name=agent_name)
        self.register_entry(Entry.GET, build_mock_url(agent_uri), {"name": agent_name})

        invoke_uri = Agent.URIS['invoke_service'].format(agent_name=agent_name, service_name=service_name)
        self.register_entry(Entry.POST, build_mock_url(invoke_uri), activation)

        activation_uri = ServiceActivation.URIS['activation'].format(activation_id=activation_id)
        self.register_entry(Entry.GET, build_mock_url(activation_uri), activation_msg)

        agent = self.cortex.agent(agent_name)
        self.assertEqual(agent.name, agent_name)

        service_activation = agent.invoke_service_async(service_name, self.cortex.message({'text': 'Testing...'}))
        self.assertEqual(service_activation._activation_id, activation_id)



'''
functions for mocking connection to cortex for testing 
'''


import json

from mocket.mockhttp import Entry

def john_doe_token():
    '''token with user name John Doe, this is a
        slight modification to the default at
        https://jwt.io/ debugger, with this json
        payload:
            {
            "sub": "mliu",
            "tenant": "mliu",
            "name": "John Doe",
            "iat": 1516239022,
            "exp": 12312411251
            }
    '''
    return 'eyJhbGciOiJIUzM4NCIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJtbGl1IiwidGVuYW50IjoibWxpdSIsIm5hbWUiOiJKb2huIERvZSIsImlhdCI6MTUxNjIzOTAyMiwiZXhwIjoxMjMxMjQxMTI1MX0.lUathguAHKBxAlrN23ApExCB1b5OLf37EfX1gjI_plsPYYMyE0jHPUguByqrwKxT'

def john_doe_subject():
    '''
    The subject part of the jwt token
    '''
    return 'mliu'


def build_mock_url(uri, version=3):
    '''
    build a mock url for testing
    '''
    return "{api_endpoint}/v{version}/{uri}".format(api_endpoint=mock_api_endpoint(), version=version, uri=uri)


def mock_api_endpoint():
    '''
    the url endppoint for mocking
    '''
    return 'http://1.2.3.4:8000'


def register_entry(verb, url, body:dict):
    print('Registering mock for', verb, url)
    Entry.single_register(verb, url, status=200, body=json.dumps(body))


def register_entry_from_path(verb, url, path:str):
    with open(path) as fh:
        register_entry(verb, url, json.load(fh))



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from cortex.catalog import CatalogClient
from cortex.serviceconnector import ServiceConnector
from cortex.utils import get_logger
from .camel import CamelResource
from .utils import raise_for_status_with_detail

log = get_logger(__name__)


class Schema(CamelResource):
    """
    Represents a plan for accessing a service.
    """
    def __init__(self, schema, connector: ServiceConnector):
        super().__init__(schema, True)
        self._connector = connector

    @staticmethod
    def get_schema(name, client: CatalogClient):
        """
        Fetches a schema to work with.

        :param client: The client instance to use.
        :param name: The name of the schema to retrieve.
        :return: A schema object.
        """
        uri = 'catalog/types/{name}'.format(name=name)
        log.debug('Getting schema using URI: %s' % uri)
        r = client._serviceconnector.request('GET', uri)
        raise_for_status_with_detail(r)

        return Schema(r.json(), client._serviceconnector)



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import matplotlib.pyplot as plt
import seaborn as sns


class Viz:

    """
    A wrapper that assists users in running multiple commonly used visualizations on the same dataframe; not meant to be directly instantiated by clients.
    """

    def __init__(self, df, figsize=(18, 9)):
        self.df = df
        self.corr_m = df.corr()

        if figsize is None or not isinstance(figsize, tuple):
            self.figsize = (18, 9)
        else:
            self.figsize = figsize

    def show_corr(self, column: str):
        cm_sorted = self.corr_m[column].sort_values(ascending=False)
        plt.rcParams['figure.figsize'] = self.figsize
        plt.xticks(rotation=90)
        sns.barplot(x=cm_sorted.index, y=cm_sorted)
        plt.show()
        plt.clf()

    def show_corr_heatmap(self, **kwargs):
        plt.rcParams['figure.figsize'] = self.figsize
        sns.heatmap(self.corr_m, annot=True, cmap=kwargs.get('cmap', 'BuGn'), robust=True, fmt=kwargs.get('fmt', '.1f'))
        plt.show()
        plt.clf()

    def show_dist(self, column: str):
        try:
            from scipy.stats import norm
            fit = norm
        except ImportError:
            fit = None

        plt.rcParams['figure.figsize'] = self.figsize
        sns.distplot(self.df[column], fit=fit)
        plt.show()
        plt.clf()

    def show_probplot(self, column: str):
        try:
            from scipy import stats
            plt.rcParams['figure.figsize'] = self.figsize
            stats.probplot(self.df[column], plot=plt)
            plt.show()
            plt.clf()
        except ImportError:
            raise Exception('show_probplot requires SciPy to be installed')

    def show_corr_pairs(self, column: str, threshold=0.7):
        cm = self.df.corr()
        values = list(cm[column].values)
        keys = list(cm[column].keys())
        vars = [i for i in keys if values[keys.index(i)] > threshold]

        plt.rcParams['figure.figsize'] = self.figsize
        sns.pairplot(self.df, height=3, vars=vars)
        plt.show()
        plt.clf()


"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest
import json
import os

from cortex import Cortex
from cortex.experiment import ExperimentClient

from mocket.mockhttp import Entry
from mocket import mocketize
from .fixtures import john_doe_token, build_mock_url, mock_api_endpoint
import dill


class TestExperiment(unittest.TestCase):
    '''
    Test experiment checks experiment functionality
    '''    
    # values for testing experiments
    EXP_NAME = 'unittest/exp'

    def setUp(self):
        self.local = Cortex.local()
        self.local_tmp = Cortex.local('/tmp/cortex')
        self.cortex = Cortex.client(api_endpoint=mock_api_endpoint(), api_version=3, token=john_doe_token())

        # register mock for getting an expeiment from the client
        uri = ExperimentClient.URIs['experiment'].format(experiment_name=self.EXP_NAME)
        returns = {"name": self.EXP_NAME}
        Entry.single_register(Entry.GET, build_mock_url(uri), status=200, body=json.dumps(returns))

    @mocketize
    def test_make_remote_experiment(self):       
        exp = self.cortex.experiment(self.EXP_NAME, version='3')
        self.assertNotEqual(exp, None)

    def test_make_local_experiment(self):
        exp = self.local.experiment(self.EXP_NAME)
        self.assertNotEqual(exp, None)

    def test_make_local_experiment_custom_basedir(self):
        exp = self.local_tmp.experiment(self.EXP_NAME)
        self.assertNotEqual(exp, None)
        self.assertTrue(os.path.isdir(f'/tmp/cortex/local/experiments/{self.EXP_NAME}'))

    @mocketize
    def test_remote_load_artifact(self):
        exp = self.cortex.experiment(self.EXP_NAME, version='3')
        # add a run & artifact

        ## register mock for creating a run
        uri = ExperimentClient.URIs['runs'].format(experiment_name=self.EXP_NAME)
        run_id = '000001'
        returns = {"runId": run_id}
        Entry.single_register(Entry.POST, build_mock_url(uri), status=200, body=json.dumps(returns))

        run = exp.start_run()
        ## make an artifact
        my_dictionary = {'a_thing': 1, 'another_thing': 2}

        # get the artifact & test if it is what is expected
        ## register mock for loading an artifact
        uri = ExperimentClient.URIs['artifact'].format(experiment_name=self.EXP_NAME, run_id=run_id, artifact='my_dictionary')
        Entry.single_register(Entry.GET, build_mock_url(uri), status=200, body=dill.dumps(my_dictionary))

        result = exp.load_artifact(run, 'my_dictionary')
        self.assertEqual(result, my_dictionary)

    def test_save_multiple_local_experiments(self):
        exp = self.local.experiment(self.EXP_NAME)
        exp.reset()
        run_a = self.make_run(exp)
        run_b = self.make_run(exp)

        self.assertTrue(len(exp.runs()) == 2)

        # does exp have runs?
        self.assertFalse(exp.get_run(run_a.id) == None)
        self.assertFalse(exp.get_run(run_b.id) == None)

        # Are runs gone after reset and retrival?
        exp.reset()

        exp = self.local.experiment(self.EXP_NAME)
        self.assertTrue(len(exp.runs()) == 0)

    def test_reset_works_in_context(self):
        exp = self.local.experiment(self.EXP_NAME)
        exp.reset()
        exp.set_meta('style', 'supervised')
        exp.set_meta('function', 'regression')
        test_artifact = {'model': {}}
        with exp.start_run() as run:
            run.log_artifact('test', test_artifact)

        self.assertTrue(len(exp.runs()) == 1)
        self.assertEqual(run.get_artifact('test'), test_artifact)

    def make_run(self, exp):
        run = exp.start_run()
        run.set_meta('meta','meta_val')
        run.log_params({'param':'param_val'})
        run.start()
        run.stop()
        run.log_metric(name='metric_1',metric=85)
        exp.save_run(run)
        return run



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import threading


class PipelineLoader:
    """
    Manages a collection of pipelines.
    """

    _default_loader = None
    _default_loader_lock = threading.Lock()

    @classmethod
    def default_loader(cls):
        with cls._default_loader_lock:
            if cls._default_loader is None:
                cls._default_loader = cls()
            return cls._default_loader

    def __init__(self):
        self._pipelines = {}

    def add_pipeline(self, name, pipeline):
        self._pipelines[name] = pipeline

    def get_pipeline(self, name):
        p = self._pipelines.get(name)
        if not p:
            from .pipeline import Pipeline
            p = Pipeline(name=name, loader=self)
            self.add_pipeline(name, p)
        return p

    def remove_pipeline(self, name):
        if name in self._pipelines:
            del self._pipelines[name]

    def dump(self, camel='1.0.0'):
        return {k: v.to_camel(camel) for k, v in self._pipelines.items()}



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json

from .serviceconnector import _Client
from .utils import get_logger
from .utils import raise_for_status_with_detail

log = get_logger(__name__)


class AuthenticationClient(_Client):
    """
    Client authentication.
    """

    URIs = {'authenticate': 'admin/{}/users/authenticate',
            'register':     'admin/tenants/register',
            'upgrade':      'accounts/token/upgrade'
           }

    def fetch_auth_token(self, tenant_id, username, password):
        """
        Retrieves the JWT token for a given user in a given tenant.

        :param tenant_id: The ID of the tenant/account to authenticate to.
        :param username: The name of the user.
        :param password: The user's password.
        :return: A JWT string.
        """
        uri = self.URIs['authenticate'].format(tenant_id)
        body = {'username': username,
                'password': password}
        body_s = json.dumps(body)
        headers = {'Content-Type': 'application/json'}
        res = self._serviceconnector.request('POST', uri, body_s, headers)
        raise_for_status_with_detail(res)
        return res.json()['jwt']

    def register(self, tenant_info, invitation_code):
        """
        Registers a client with an invitation code.

        :param tenant_info: The tenant to register.
        :param invitation_code: The invitation code for the registration requset.
        """
        uri = self.URIs['register']
        body_s = json.dumps(tenant_info)
        headers = {'Content-Type': 'application/json'}
        params = {'invitationCode': invitation_code}
        res = self._serviceconnector.request('POST', uri, body_s, headers, params=params)
        raise_for_status_with_detail(res)
        return res.json()



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from .utils import get_cortex_profile
from .exceptions import BadTokenException

DEFAULT_API_ENDPOINT = 'https://api.cortex.insights.ai'


class CortexEnv:
    """
    Sets environment variables for Cortex.
    """

    def __init__(self, api_endpoint=None, token=None, account=None, username=None, password=None):
        profile = CortexEnv.get_cortex_profile()

        cortexToken = token or CortexEnv.get_token()

        if not cortexToken:
            raise BadTokenException(
                'Your Cortex credentials cannot be retrieved. For more information, go to Cortex Docs > Cortex Tools > Access')

        self.api_endpoint = api_endpoint or os.getenv('CORTEX_URI', profile.get('url'))
        self.token = cortexToken
        self.account = account or os.getenv('CORTEX_ACCOUNT', profile.get('account'))
        self.username = username or os.getenv('CORTEX_USERNAME', profile.get('username'))
        self.password = password or os.getenv('CORTEX_PASSWORD')

    @staticmethod
    def get_token():
        """
        gets the token from either the cortex_token env variable or the profile's token.
        if cortex_token and both cortex_profile are falsey, then cortexToken will be None

        :param profile: configured cortex profile from the local machine
        """
        cortexToken = CortexEnv.get_cortex_token() or CortexEnv.get_cortex_profile().get('token')
        return cortexToken

    @staticmethod
    def get_cortex_profile():
        """
        gets the configured cortex profile from the local machine
        """
        return get_cortex_profile()

    @staticmethod
    def get_cortex_token():
        """
        gets the cortex token from the local machine
        """
        return os.getenv('CORTEX_TOKEN')



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import getpass
import os
import time

from .serviceconnector import ServiceConnector
from .auth import AuthenticationClient
from .action import Action
from .agent import Agent
from .dataset import Dataset, LocalDataset, DatasetsClient
from .env import CortexEnv
from .exceptions import AuthenticationException, ConfigurationException
from .experiment import Experiment, LocalExperiment, ExperimentClient
from .message import Message
from .session import Session, SessionClient
from .skill import Skill
from .utils import log_message, decode_JWT, get_logger

_DEFAULT_API_VERSION = 3

_msg_token_exp_no_creds = """
Your Cortex token is expired, and the required credentials for auto-refresh have not been provided. Supply these credentials; account, username,
and password.  Please login again to retrieve a valid token.
"""

log = get_logger(__name__)


class _Token(object):

    def __init__(self, auth_client: AuthenticationClient, token: str, account: str, username: str, password: str):
        self._auth = auth_client
        self._token = token
        self._account = account
        self._username = username
        self._password = password

        self._jwt = None
        if token:
            self._jwt = decode_JWT(self._token, verify=False)

    def login(self):
        try:
            log_message('Login with user %s/%s' % (self._account, self._username), log)
            self._token = self._auth.fetch_auth_token(self._account, self._username, self._password)
        except Exception as e:
            raise AuthenticationException(str(e))

        self._jwt = decode_JWT(self._token, verify=False)

    def is_expired(self):
        current_time = time.time()
        return not self._jwt or (self._jwt.get('exp', current_time) < current_time)

    @property
    def token(self):
        return self._token


class Client(object):

    """
    API client used to access agents, skills, and datasets.
    """

    def __init__(self, url: str, token: _Token, version: int = 3, verify_ssl_cert: bool = False):
        self._token = token
        self._url = url
        self._version = version
        self._verify_ssl_cert = verify_ssl_cert

    def agent(self, name: str) -> Agent:
        """
        Gets an agent with the specified name.
        """
        return Agent.get_agent(name, self._mk_connector())

    def skill(self, name: str) -> Skill:
        """
        Gets a skill with the specified name.
        """
        return Skill.get_skill(name, self._mk_connector())

    def dataset(self, name: str) -> Dataset:
        """
        Gets a dataset with the specified name.
        """
        ds_client = DatasetsClient(self._url, self._version, self._token.token)
        return Dataset.get_dataset(name, ds_client)

    def session(self, session_id=None, ttl=None, instance_id=None) -> Session:
        """
        Gets a session with the specified identifier.
        """
        session_client = SessionClient(self._url, self._version, self._token.token)
        if not session_id:
            return Session.start(session_client, ttl, instance_id)
        return Session(session_id, session_client)

    def action(self, name: str) -> Action:
        """
        Gets an action with the specified name.
        """
        return Action.get_action(name, self._mk_connector())

    def builder(self):
        """
        Gets a builder.
        """
        try:
            from cortex_builders import Builder
            return Builder(self)
        except ImportError:
            raise ConfigurationException('Please install the cortex-python-builders library to use this function')

    def experiment(self, name: str, version: str = '2'):
        """
        Gets an experiment with the specified name.
        """
        exp_client = ExperimentClient(self._url, version, self._token.token)
        return Experiment.get_experiment(name, exp_client)

    def message(self, payload: dict, properties: dict = None) -> Message:
        """Constructs a Message from payload and properties if given.

        :param payload: The payload to include in the Message.
        :param properties: The properties to include in the Message.
        :return: A Message object.
        """
        params = {}
        params['payload'] = payload
        if properties:
            params['properties'] = properties
        params['apiEndpoint'] = self._url
        params['token'] = self._token.token
        return Message(params)

    def _mk_connector(self):
        return ServiceConnector(self._url, self._version, self._token.token, self._verify_ssl_cert)

    # expose this to allow developer to pass client instance into Connectors
    def to_connector(self):
        return self._mk_connector()

class Local:

    """
    Provides local, on-disk implementations of Cortex APIs.
    """

    def __init__(self, basedir=None):
        self._basedir = basedir

    def experiment(self, name: str) -> LocalExperiment:
        return LocalExperiment(name, self._basedir)

    def dataset(self, name: str):
        return LocalDataset(name, self._basedir)

    def action(self, name: str):
        raise NotImplementedError('Local actions are not implemented yet.')

    def builder(self):
        try:
            from cortex_builders import LocalBuilder
            return LocalBuilder(self._basedir)
        except ImportError:
            raise ConfigurationException('Please install the cortex-python-builders library to use this function')


class Cortex(object):

    """
    Entry point to the Cortex API.
    """

    @staticmethod
    def client(api_endpoint:str=None, api_version:int=_DEFAULT_API_VERSION, verify_ssl_cert=None, token:str=None, account:str=None, username:str=None, password:str=None):
        """
        Gets a client with the provided parameters. All parameters are optional and default to environment variable values if not specified.

        **Example**

        >>> from cortex import Cortex
        >>> cortex = Cortex.client()

        :param api_endpoint: The Cortex URL.
        :param api_version: The version of the API to use with this client.
        :param verify_ssl_cert: A boolean to indiciate if the SSL certificate needs to be validated.
        :param token: An authentication token.
        :param account: The Cortex account the client connects to.
        :param username: The user name for the client.
        :param password: The password for the client.
        """
        env = CortexEnv(api_endpoint=api_endpoint, token=token, account=account, username=username, password=password)

        if not api_endpoint:
            api_endpoint = env.api_endpoint

        if not token:
            token = env.token

        if not account:
            account = env.account

        if not username:
            username = env.username

        if not password:
            password = env.password

        auth = AuthenticationClient(api_endpoint, version=2, verify_ssl_cert=verify_ssl_cert)
        t = _Token(auth, token, account, username, password)

        return Client(url=api_endpoint, version=api_version, token=t, verify_ssl_cert=verify_ssl_cert)

    @staticmethod
    def from_message(msg):
        """
        Creates a Cortex client from a message that must incluide an API endpoint and a token.

        :param msg: A message for constructing a Cortex Client.
        """
        return Cortex.client(api_endpoint=msg.apiEndpoint, token=msg.token)

    @staticmethod
    def local(basedir=None):
        return Local(basedir)

    @staticmethod
    def login():
        """
        Login to Cortex5. The function prompts the caller for Cortex URI, account, user name, and password, retrieves the user's JWT token and sets the information as environment variables for use by the SDK.

        **Example**

        >>> Cortex.login()
        Cortex URI: [https://api.cortex.insights.ai]
        Account: accountId
        Username: userName
        Password: ****
        """
        prod_uri = 'https://api.cortex.insights.ai'
        uri = input('Cortex URI: [{}]'.format(prod_uri)) or prod_uri
        account = input('Account: ')
        username = input('Username: ')
        password = getpass.getpass('Password: ')
        auth = AuthenticationClient(uri, version=2)
        token = auth.fetch_auth_token(account, username, password)
        os.environ['CORTEX_URI'] = uri
        os.environ['CORTEX_ACCOUNT'] = account
        os.environ['CORTEX_TOKEN'] = token
        os.environ['CORTEX_USERNAME'] = username
        os.environ['CORTEX_PASSWORD'] = password



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import yaml
from datetime import datetime
import os
from contextlib import closing
import threading
import copy
from pathlib import Path


class PropertyManager:
    """
    Maintains a list of key-value pairs.
    """
    def __init__(self):
        self._properties = dict({})

    def is_key(self, key) -> bool:
        """
        Identifies if a key exists or not.

        :param key: The key of the value; should be a dot separated string of keys from root up the tree.
        :return: `True` if the key exists in the properties; `False` if the key doesn't exist in the properties.
        """
        if key is None or not key:
            return False
        find_dict = self._properties
        is_path, _, is_key = key.rpartition('.')
        if is_path:
            for part in is_path.split('.'):
                if isinstance(find_dict, dict):
                    find_dict = find_dict.get(part, {})
                else:
                    break
        if is_key in find_dict:
            return True
        return False

    def save(self, file_name: str):
        """
        Saves properties to a file.

        :param file_name: file in which to save properties
        """
        _time = str(datetime.now())
        self.set('meta', {'updated': _time})
        self._dump(self._properties, file_name)

    def load(self, config_file, replace=False) -> bool:
        """
        Loads the properties from the YAML configuration file; allows for multiple configuration.

        Files to be merged into the properties dictionary, or properties to be refreshed in real time.

        :param config_file: the path and filename of the YAML file
        :param replace: boolean value that identifies if the file is to be added or replaced
        :return: `True` if the file is found, opened, and loaded; `False` if an exception occurs

        """
        if config_file is None or not config_file or not isinstance(config_file, str):
            raise ValueError("The properties configuration file must be a valid str")

        _path = Path(config_file)
        if not _path.exists() or not _path.is_file():
            raise FileNotFoundError("Can't find the configuration file {}".format(_path))

        try:
            cfg_dict = self._load(_path)
        except (IOError, TypeError) as e:
            raise IOError(e)
        if replace:
            with threading.Lock():
                self._properties.clear()
                self._properties = cfg_dict
            return True
        for _key, _value in cfg_dict.items():
            # only replace sections that have changed
            if self.is_key(_key) and cfg_dict.get(_key) == self.get(_key):
                continue
            self.remove(_key)
            self.set(_key, _value)
        return True

    def get(self, key: str):
        """
        Gets a property value for the dot-separated key.

        :param key: the key of the value; should be a dot-separated string of keys from root up the tree

        :return:
            An object found in the key can be any structure found under that key.
            If the key is not found, `None` is returned.
            If the key is `None`, then the complete properties dictionary is returned.
            The full tree is returned under the requested key, be it a value, tuple, list, or dictionary.
        """
        if key is None or not key:
            return None

        rtn_val = self._properties
        for part in key.split('.'):
            if isinstance(rtn_val, dict):
                rtn_val = rtn_val.get(part)
                if rtn_val is None:
                    return None
            else:
                return None
        with threading.Lock():
            return copy.deepcopy(rtn_val)

    def set(self, key: str, value):
        """
        Sets a key-value pair. The value can be a union (str, dict, tuple, array).

       :param key: the key string
       :param value: the value of the key
       """
        if key is None or not key or not isinstance(key, str):
            raise ValueError("The key must be a valid str")
        keys = key.split('.')
        _prop_branch = self._properties
        _last_key = None
        _last_prop_branch = None
        # from base of the key work up to find where the section doesn't exist
        for _, k in list(enumerate(keys, start=0)):
            if k not in _prop_branch:
                break
            _last_prop_branch = _prop_branch
            _last_key = k
            _prop_branch = _prop_branch[k]
        tmp_dict = {}
        # now from the top of the key work back, creating the sections tree
        k = None
        for _, k in reversed(list(enumerate(keys, start=0))):
            if isinstance(value, dict):
                tmp_dict = {k: value}
            else:
                tmp_dict[k] = value
            if k is _last_key:
                break
            value = tmp_dict
        if not isinstance(value, dict):
            _last_prop_branch[k] = value
            return
        self._add_value(k, value, _prop_branch)
        return

    def _add_value(self, key, value, base):
        if key is None:
            return None
        for k, v in value.items():
            if isinstance(v, dict):
                if k in base:
                    base = base[k]
                    self._add_value(k, v, base)
                else:
                    with threading.Lock():
                        base.update({k: v})
            else:
                with threading.Lock():
                    base[k] = v
        return

    def remove(self, key) -> bool:
        """
        Removes a key-value from the in-memory configuration dictionary based on the key.

        :param key: the key of the key/value to be removed;
            should be a dot-separated string of keys from root up the tree

        :return: `True` if the key is removed; `False` if the key is not found
        """
        del_dict = self._properties
        del_path, _, del_key = key.rpartition('.')
        if del_path:
            for part in del_path.split('.'):
                if isinstance(del_dict, dict):
                    del_dict = del_dict.get(part)
                else:
                    return False
        if del_dict is None or del_key not in del_dict:
            return False
        with threading.Lock():
            del del_dict[del_key]
        return True

    def remove_all(self):
        with threading.Lock():
            meta = self._properties.get('meta')
            self._properties.clear()
            self.set('meta', meta)

    def get_all(self) -> dict:
        """
        Gets all the properties.

        :returns: a deep copy of the key-value pairs
        """
        with threading.Lock():
            return copy.deepcopy(self._properties)

    @staticmethod
    def join(*names, sep=None) -> str:
        """
        Used to create a name string or join paths by passing sep=os.path.sep.

        :param names: the names to join
        :param sep: the join separator; defaults to '.'
        :return: the names joined with the separator
        """
        _sep = sep if sep is not None else '.'
        return _sep.join(map(str, names))

    @staticmethod
    def _dump(data, config_file) -> None:
        _path, _file = os.path.split(config_file)
        if not os.path.exists(_path):
            os.makedirs(_path, exist_ok=True)
        _config_file = Path(config_file)
        with threading.Lock():
            # make sure the dump is clean
            try:
                with closing(open(_config_file, 'w')) as ymlfile:
                    yaml.safe_dump(data, ymlfile, default_flow_style=False)
            except IOError as e:
                raise IOError("The configuration file {} failed to open with: {}".format(config_file, e))
        # check the file was created
        if not _config_file.exists():
            raise IOError("Failed to save configconfiguration file {}. Check the disk is writable".format(config_file))
        return

    @staticmethod
    def _load(config_file) -> dict:
        config_file = Path(config_file)
        if not config_file.exists():
            raise FileNotFoundError("The configuration file {} does not exist".format(config_file))
        with threading.Lock():
            try:
                with closing(open(config_file, 'r')) as ymlfile:
                    rtn_dict = yaml.safe_load(ymlfile)
            except IOError as e:
                raise IOError("The configuration file {} failed to open with: {}".format(config_file, e))
            if not isinstance(rtn_dict, dict) or not rtn_dict:
                raise TypeError("The configuration file {} could not be loaded as a dict type".format(config_file))
            return rtn_dict



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from io import BytesIO
import json
import unittest

from mocket.mockhttp import Entry
from mocket import mocketize

from cortex.connection import ConnectionClient
from cortex.content import ManagedContentClient


class TestConnectionClient(unittest.TestCase):

    def setUp(self):
        self.cc = ConnectionClient('http://localhost:123', 2, 'token')
        self.mc = ManagedContentClient('http://localhost:123', 2, 'token')

    @mocketize
    def test_save_connection(self):
        uri = 'connections'
        url = self.cc._serviceconnector._construct_url(uri)
        connection = { 'connectionType': 'ctype', 'name': 'cname' }
        Entry.single_register(Entry.POST,
                              url,
                              status = 200,
                              body = json.dumps(connection))
        r = self.cc.save_connection(connection)

        self.assertEqual(r, connection)

    @mocketize
    def test_upload(self):
        key = 'some-key'
        result = { 'Key': key }
        uri = 'content'
        url = self.cc._serviceconnector._construct_url(uri)

        with BytesIO(b'arbitrary content') as content:
            Entry.single_register(Entry.POST,
                                  url,
                                  status = 200,
                                  body = json.dumps(result))
            r = self.mc.upload(key=key, stream_name='foo', stream=content, content_type='application/octet-stream')
            self.assertEqual(r, result)

    @mocketize
    def test_uploadStreaming(self):
        key = 'some-key'
        result = { 'Key': key }
        uri = 'content/{0}'.format(key)
        url = self.cc._serviceconnector._construct_url(uri)

        with BytesIO(b'arbitrary content') as content:
            Entry.single_register(Entry.POST,
                                  url,
                                  status = 200,
                                  body = json.dumps(result))
            r = self.mc.upload_streaming(key=key, stream=content, content_type='application/octet-stream')
            self.assertEqual(r, result)

    @mocketize
    def test_download(self):
        key = 'some-key'
        uri = 'content/{0}'.format(key)
        url = self.cc._serviceconnector._construct_url(uri)

        buf = b'arbitrary content'
        with BytesIO(buf) as content:
            Entry.single_register(Entry.GET,
                                url,
                                status = 200,
                                body = content)
            r = self.mc.download(key=key)
            self.assertEqual(r.read(), buf)

    @mocketize
    def test_exists(self):
        key = 'some-key'
        result = { 'Key': key }
        uri = 'content/{0}'.format(key)
        url = self.cc._serviceconnector._construct_url(uri)
        Entry.single_register(Entry.HEAD,
                              url,
                              status = 200,
                              body = json.dumps(result))
        
        r = self.mc.exists(key=key)
        self.assertTrue(r)



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict

from .utils import get_logger
from .camel import Document
from .env import CortexEnv

log = get_logger(__name__)


class Message(Document):
    """
    Wraps a set of parameters or a payload when invoking an action, skill, or agent.

    params:
    instanceId: The ID of the agent instance that produced the message.
    typeName: The resource name of the type/schema contained in the payload of this message.
    token: The JWT token to be used for making authenticated requests to APIs needed to process this message.
    sessionId: The ID of the session associated with this message
    channelId: The ID of the channel (wire) this message was sent on
    apiEndpoint: The URI of the API server to use when making requests to platform services in processing of this message.
    properties: An object whose keys represent metadata properties set for the message.
    payload: The body of the message in one of three supported payload formats (see the CAMEL spec for details).
    """

    def __init__(self, params:Dict = None):
        if params is None:
            params = {}

        super().__init__(params, False)
        self._params = params

    def to_params(self) -> Dict:
        """
        Gets the parameters in the message.
        """
        return self._params

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

        if not key.startswith('_'):
            self._params[key] = value

    def as_pandas(self, client=None):
        """
        Returns the message's payload as a pandas dataframe
        """
        if not hasattr(self, 'payload') or self.payload is None:
            raise AttributeError('Message is missing payload attribute')

        columns = []
        values = []

        ds_ref = self.payload.get('$ref')
        if ds_ref:
            if client is None:
                from .client import Cortex
                client = Cortex.from_message(self)

            ds = client.dataset(ds_ref)
            df = ds.get_dataframe()
            columns = df.get('columns')
            values = df.get('values')

        elif 'records' in self.payload:
            records = self.payload.get('records')
            if len(records) > 0:
                columns = records[0].keys()
                for obj in records:
                    values.append([obj[key] for key in columns])

        else:
            values = self.payload.get('values')
            if not values:
                raise ValueError('Invalid DataFrame: values missing from payload')

            columns = self.payload.get('columns')
            if not columns:
                raise ValueError('Invalid DataFrame: columns missing from payload')

        try:
            import pandas as pd
            return pd.DataFrame(values, columns=columns)
        except ImportError:
            log.warn('Pandas is not installed, please run `pip install pandas` or equivalent in your environment')
            return {'columns': columns, 'values': values}

    def get_dataset(self):
        """
        Returns the message's payload as a dataset.
        """
        if not hasattr(self, 'payload') or self.payload is None:
            raise AttributeError('Message is missing payload attribute')

        ds_ref = self.payload.get('$ref')
        if ds_ref:
            from .client import Cortex
            client = Cortex.from_message(self)
            return client.dataset(ds_ref)

        raise AttributeError('Message payload does not contain a dataset reference')

    @staticmethod
    def from_env(**kwargs):
        env = CortexEnv(**kwargs)

        params = {}
        if env.api_endpoint:
            params['apiEndpoint'] = env.api_endpoint
        if env.token:
            params['token'] = env.token

        return Message(params)



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest

from unittest.mock import Mock

from cortex.env import CortexEnv
from cortex.exceptions import BadTokenException


class TestCortexConfiguration(unittest.TestCase):

    def setUp(self):
        self.getCortexTokenOriginal = CortexEnv.get_cortex_token
        self.getCortexProfileOriginal = CortexEnv.get_cortex_profile

    def test_get_cortex_token(self):
        token = 'some_token'
        CortexEnv.get_cortex_token = Mock(return_value=token)
        self.assertEqual(CortexEnv.get_cortex_token(), token)

    def test_get_cortex_profile(self):
        profile = { "A" : 1, "B": 2 } 
        CortexEnv.get_cortex_profile = Mock(return_value=profile)
        self.assertEqual(CortexEnv.get_cortex_profile(), profile)
    
    def test_get_token_with_token_and_profile(self):
        token = 'some_token'
        profile =  {
            "url": "https://api.cortex-dev.insights.ai",
            "username": "adminTenant-666",
            "account": "testtenant-666",
            "token": "123c3MiOiJjb2duaXRpdmVzY2FsZS5jb20iLCJhdWQiOiJjb3J0ZXgiLCJzdWIiOiJhZG1pblRlbmFudC02NjYiLCJ0ZW5hbnQiOiJ0ZXN0dGVuYW50LTY2NiIsImJlYXJlciI6InB1YmxpYyIsImtleSI6Im5nM2NTa2Y0R2FlSmticU9wN1BIbVRNVHFPekh5eEc2IiwiZXhwIjoxNTU4MzY3ODEyLCJhY2wiOnsiLioiOlsiUkVBRCIsIlJVTiIsIldSSVRFIiwiREVMRVRFIl19LCJpYXQiOjE1NTcxNTgyMTJ9.VQPxm6j2i2QzqFGQGw-dMSSU8TkxQwkx_e9WN1tdHcU"
        }
        CortexEnv.get_cortex_token = Mock(return_value=token)
        CortexEnv.get_cortex_profile = Mock(return_value=profile)
        self.assertEqual(CortexEnv.get_token(), token)

    def test_get_token_with_token_and_empty_profile(self):
        token = 'some_token'
        profile = {}
        CortexEnv.get_cortex_token = Mock(return_value=token)
        CortexEnv.get_cortex_profile = Mock(return_value=profile)
        self.assertEqual(CortexEnv.get_token(), token)
    
    def test_get_token_with_no_token_and_profile(self):
        profile_token = '123c3MiOiJjb2duaXRpdmVzY2FsZS5jb20iLCJhdWQiOiJjb3J0ZXgiLCJzdWIiOiJhZG1pblRlbmFudC02NjYiLCJ0ZW5hbnQiOiJ0ZXN0dGVuYW50LTY2NiIsImJlYXJlciI6InB1YmxpYyIsImtleSI6Im5nM2NTa2Y0R2FlSmticU9wN1BIbVRNVHFPekh5eEc2IiwiZXhwIjoxNTU4MzY3ODEyLCJhY2wiOnsiLioiOlsiUkVBRCIsIlJVTiIsIldSSVRFIiwiREVMRVRFIl19LCJpYXQiOjE1NTcxNTgyMTJ9.VQPxm6j2i2QzqFGQGw-dMSSU8TkxQwkx_e9WN1tdHcU'
        profile =  {
            "url": "https://api.cortex-dev.insights.ai",
            "username": "adminTenant-666",
            "account": "testtenant-666",
            "token": profile_token
        }
        CortexEnv.get_cortex_token = Mock(return_value=None) # notice this is None (not the profile_token)
        CortexEnv.get_cortex_profile = Mock(return_value=profile)
        self.assertEqual(CortexEnv.get_token(), profile_token)

    def test_get_token_with_no_token_no_profile(self):
        profile_token = ''
        profile =  {}
        CortexEnv.get_cortex_token = Mock(return_value=profile_token)
        CortexEnv.get_cortex_profile = Mock(return_value=profile)
        self.assertEqual(CortexEnv.get_token(), None)

    def test_get_token_with_token_and_profile2(self):
        token = '123c3MiOiJjb2duaXRpdmVzY2FsZS5jb20iLCJhdWQiOiJjb3J0ZXgiLCJzdWIiOiJhZG1pblRlbmFudC02NjYiLCJ0ZW5hbnQiOiJ0ZXN0dGVuYW50LTY2NiIsImJlYXJlciI6InB1YmxpYyIsImtleSI6Im5nM2NTa2Y0R2FlSmticU9wN1BIbVRNVHFPekh5eEc2IiwiZXhwIjoxNTU4MzY3ODEyLCJhY2wiOnsiLioiOlsiUkVBRCIsIlJVTiIsIldSSVRFIiwiREVMRVRFIl19LCJpYXQiOjE1NTcxNTgyMTJ9.VQPxm6j2i2QzqFGQGw-dMSSU8TkxQwkx_e9WN1tdHcU'
        profile =  {
            "url": "https://api.cortex-dev.insights.ai",
            "username": "adminTenant-666",
            "account": "testtenant-666",
            "token": 'mumjobotoken'
        }
        CortexEnv.get_cortex_token = Mock(return_value=token) # notice this is None
        CortexEnv.get_cortex_profile = Mock(return_value=profile)
        self.assertEqual(CortexEnv.get_token(), token)

    def test_constructor_no_profile_and_no_token(self):
        token = ''
        profile = {}
        CortexEnv.get_cortex_token = Mock(return_value=token)
        CortexEnv.get_cortex_profile = Mock(return_value=profile)
        self.assertRaises(BadTokenException, CortexEnv)

    # we don't want methods calls to CortexEnv to use the monkey patched methods
    # so we revert back to the original methods.
    def tearDown(self):
        CortexEnv.get_cortex_token = self.getCortexTokenOriginal
        CortexEnv.get_cortex_profile = self.getCortexProfileOriginal



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import unittest
import uuid

from mocket.mockhttp import Entry
from mocket import mocketize

from cortex.session import SessionClient


class TestSessionClient(unittest.TestCase):

    def setUp(self):
        self.client = SessionClient('http://localhost:8000', 3, 'token')
        self.session_id = str(uuid.uuid4())

    def register_entry(self, verb, uri, body):
        url = self.client._serviceconnector._construct_url(uri)
        Entry.single_register(verb,
                              url,
                              status=200,
                              body=json.dumps(body))

    ## Sessions ##

    @mocketize
    def test_start_sessions(self):
        uri = self.client.URIs['start']
        returns = {'sessionId': self.session_id}
        self.register_entry(Entry.POST, uri, returns)

        r = self.client.start_session()
        self.assertEqual(r, self.session_id)

    @mocketize
    def test_get_all(self):
        uri = self.client.URIs['get'].format(session_id=self.session_id)
        returns = {'state': {'key1': 'value1'}}
        self.register_entry(Entry.GET, uri, returns)

        r = self.client.get_session_data(self.session_id)
        self.assertEqual(r, returns['state'])

    @mocketize
    def test_get_by_key(self):
        uri = self.client.URIs['get'].format(session_id=self.session_id) + '?key=key1'
        returns = {'state': {'key1': 'value1'}}
        self.register_entry(Entry.GET, uri, returns)

        r = self.client.get_session_data(self.session_id, 'key1')
        self.assertEqual(r, returns['state'])

    @mocketize
    def test_put(self):
        uri = self.client.URIs['put'].format(session_id=self.session_id)
        returns = {}
        self.register_entry(Entry.POST, uri, returns)

        r = self.client.put_session_data(self.session_id, {})
        self.assertEqual(r, returns)

    @mocketize
    def test_delete(self):
        uri = self.client.URIs['delete'].format(session_id=self.session_id)
        returns = {}
        self.register_entry(Entry.DELETE, uri, returns)

        r = self.client.delete_session(self.session_id)
        self.assertEqual(r, returns)



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict


class Document:
    """
    Defines document (read-only) attributes.
    """
    def __init__(self, document: Dict, read_only=True):
        super().__setattr__('_document', document or {})
        super().__setattr__('_read_only', read_only)

    def __getattr__(self, name):
        if name.startswith('_'):
            return super().__getattribute__(name)

        doc = super().__getattribute__('_document')
        return doc.get(name)

    def __setattr__(self, name: str, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        elif super().__getattribute__('_read_only'):
            raise AttributeError(
                'Attempt to modify a read-only attribute: %s' % name)
        else:
            doc = super().__getattribute__('_document')
            doc[name] = value


class CamelResource(Document):
    """
    Contains CAMEL attributes for a Cortex object.
    """
    def __init__(self, document: Dict, read_only=True):
        super().__init__(document, read_only)

    @property
    def name(self):
        """
        The CAMEL resource name attribute.
        """
        return self._document.get('name')

    @property
    def version(self):
        """
        The CAMEL resource version (spelled _version) attribute.
        """
        return self._document.get('_version')

    @property
    def title(self):
        """
        The CAMEL resource title attribute.
        """
        return self._document.get('title')




"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
from .serviceconnector import _Client, ServiceConnector
from .camel import CamelResource
from .utils import get_logger
from .utils import raise_for_status_with_detail

log = get_logger(__name__)


class ConnectionClient(_Client):
    """
    A client used to manage connections.
    """
    URIs = {'connections': 'connections'}

    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Content apis only have v2..
        self._serviceconnector.version = 2

    def save_connection(self, connection: object):
        """
        Posts the connection client information.
        """
        uri  = self.URIs['connections']
        data = json.dumps(connection)
        headers = {'Content-Type': 'application/json'}
        r = self._serviceconnector.request('POST', uri, data, headers)
        raise_for_status_with_detail(r)
        return r.json()

    ## Private ##

    def _bootstrap(self):
        uri  = self.URIs['connections'] + '/_/bootstrap'
        r = self._serviceconnector.request('GET', uri)
        raise_for_status_with_detail(r)
        return r.json()


class Connection(CamelResource):

    """
    Defines the connection for a dataset.
    """

    def __init__(self, connection, connector: ServiceConnector):
        super().__init__(connection, True)
        self._connector = connector

    @staticmethod
    def get_connection(name, client: ConnectionClient):
        """
        Fetches a Connection to work with.

        :param client: The client instance to use.
        :param name: The name of the connection to retrieve.
        :return: A Connection object.
        """
        uri = 'connections/{name}'.format(name=name)
        log.debug('Getting connection using URI: %s' % uri)
        r = client._serviceconnector.request('GET', uri)
        raise_for_status_with_detail(r)

        return Connection(r.json(), client._serviceconnector)



"""
Copyright 2018 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest

from cortex import Cortex
from cortex.message import Message

from .fixtures import john_doe_subject, john_doe_token


class Test_Cortex(unittest.TestCase):
    
    def test_client(self):
        api_endpoint = 'https://api.test.cortex'
        api_version = 1
        account = 'unittest'
        username = 'unittest'
        password = 'unittest'
        token = john_doe_token()

        cortex = Cortex.client(
                api_endpoint=api_endpoint,
                api_version=api_version,
                account=account,
                username=username,
                password=password,
                token=token
                )
        assert cortex._url == api_endpoint
        assert cortex._token._account == account
        assert cortex._token._username == username
        assert cortex._token._password == password
        assert cortex._token._token == token
        assert cortex._token._jwt['sub'] == john_doe_subject()

    def test_message_constructor(self):
        cortex = Cortex.client(
            api_endpoint = 'https://api.test.cortex',
            api_version = 1,
            account = 'unittest',
            username = 'unittest',
            password = 'unittest',
            token = john_doe_token()
            )
        message = cortex.message({'foo': 'bar'})
        assert isinstance(message, Message)
        assert message.apiEndpoint == cortex._url
        assert message.token == cortex._token.token
        assert message.token == john_doe_token()





"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import pkg_resources
import platform
import requests
import sys
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from typing import Dict, Any, List, Union, Optional, Type, TypeVar
from .utils import get_logger, get_cortex_profile
from .utils import raise_for_status_with_detail
log = get_logger(__name__)

JSONType = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]
T = TypeVar('T', bound="_Client")

userAgent = f'cortex-python/{pkg_resources.get_distribution("cortex-python").version} ({sys.platform}; {platform.architecture()[0]}; {platform.release()})'

class ServiceConnector:
    """
    Defines the settings and security credentials required to access a service.
    """
    def __init__(self, url, version=3, token=None, verify_ssl_cert=True):
        self.url = url
        self.version = version
        self.token = token
        self.verify_ssl_cert = verify_ssl_cert

    ## properties ##

    @property
    def base_url(self):
        return u'{0}/v{1}'.format(self.url, self.version)

    ## methods ##

    def post_file(self, uri, files, data, headers=None):
        """
        Posts to a service, extending the path with the specified URI.

        :param uri: Path to extend the service URL.
        :param files: Files to send to the service.
        :param data: Data to send as the post body to the service.
        :param headers: HTTP headers for the post.
        """
        headersToSend = self._construct_headers(headers)
        url = self._construct_url(uri)
        return requests.post(url, files=files, data=data, headers=headersToSend,
                             verify=self.verify_ssl_cert)

    def request_with_retry(self, method, uri, body=None, headers=None, debug=False, **kwargs):
        """
        Sends a request to the specified URI. Auto retry with backoff on 50x errors

        :param method: HTTP method to send to the service.
        :param uri: Path to extend service URL.
        :param data: Data to send as the post body to the service.
        :param headers: HTTP headers for this post.
        :param kwargs: Additional key-value pairs to pass to the request method.
        :return: :class:`Response <Response>` object
        """
        headersToSend = self._construct_headers(headers)
        url = self._construct_url(uri)
        if debug:
            log.debug("START {} {}".format('GET', uri))
        r = ServiceConnector.requests_retry_session().request(
            method,
            url,
            data=body,
            headers=headersToSend,
            verify=self.verify_ssl_cert,
            **kwargs
        )
        if debug:
            log.debug("  END {} {}".format('GET', uri))
        return r

    def request(self, method, uri, body=None, headers=None, debug=False, **kwargs):
        """
        Sends a request to the specified URI.

        :param method: HTTP method to send to the service.
        :param uri: Path to extend service URL.
        :param data: Data to send as the post body to the service.
        :param headers: HTTP headers for this post.
        :param kwargs: Additional key-value pairs to pass to the request method.
        :return: :class:`Response <Response>` object
        """
        headersToSend = self._construct_headers(headers)
        url = self._construct_url(uri)
        if debug:
            log.debug("START {} {}".format('GET', uri))
        r = requests.request(
            method,
            url,
            data=body,
            headers=headersToSend,
            verify=self.verify_ssl_cert,
            **kwargs
        )
        if debug:
            log.debug("  END {} {}".format('GET', uri))
        return r

    @staticmethod
    def requests_retry_session(
            retries=5,
            backoff_factor=0.5,
            status_forcelist=(500, 502, 503, 504),
            session=None,
    ):
        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    @staticmethod
    def urljoin(pieces):
        """
        Joins together the pieces of a URL.

        :parma pieces: Strings representing the pieces of a URL.
        :return: A string representing the joined pieces of the URL.
        """
        pieces = [_f for _f in [s.rstrip('/') for s in pieces] if _f]
        return '/'.join(pieces)

    ## private ##

    def _construct_url(self, uri):
        return self.urljoin([self.base_url, uri])

    def _construct_headers(self, headers):
        headersToSend = { 'User-Agent': userAgent }

        if self.token:
            auth = 'Bearer {}'.format(self.token)
            headersToSend[u'Authorization'] = auth

        if headers is not None:
            headersToSend.update(headers)
        return headersToSend


class _Client:
    """
    A client.
    """

    # type: Dict[str, str]
    URIs = {}

    def __init__(self,  *args, **kwargs):
        url = kwargs.get("url")
        version = kwargs.get("version")
        token = kwargs.get("token")
        verify_ssl_cert = kwargs.get("verify_ssl_cert")
        # If all kwargs or first arg is a string create a Connector
        if len(args) == 0 or (len(args) > 0 and type(args[0]) == str):
            if len(args) > 0:
                url = args[0]
            if len(args) > 1:
                version = args[1]
            if len(args) > 2:
                token = args[2]
            if len(args) > 3:
                verify_ssl_cert = args[3]
            self._serviceconnector = ServiceConnector(url, version, token, verify_ssl_cert)
        # if first arg not string assume Client object was passed
        else:
            self._serviceconnector = args[0].to_connector()
            self._serviceconnector.version = version

    def _post_json(self, uri, obj: JSONType):
        body_s = json.dumps(obj)
        headers = {'Content-Type': 'application/json'}
        r = self._serviceconnector.request('POST', uri, body_s, headers)
        if r.status_code not in [requests.codes.ok, requests.codes.created]:
            log.info("Status: {}, Message: {}".format(r.status_code, r.text))
        raise_for_status_with_detail(r)
        return r.json()

    def _post_json_with_retry(self, uri, obj: JSONType):
        body_s = json.dumps(obj)
        headers = {'Content-Type': 'application/json'}
        r = self._serviceconnector.request_with_retry('POST', uri, body_s, headers)
        if r.status_code not in [requests.codes.ok, requests.codes.created]:
            log.info("Status: {}, Message: {}".format(r.status_code, r.text))
        raise_for_status_with_detail(r)
        return r.json()

    def _get(self, uri, debug=False, **kwargs):
        return self._serviceconnector.request('GET', uri, debug=debug, **kwargs)

    def _delete(self, uri, debug=False):
        r = self._serviceconnector.request('DELETE', uri, debug=debug)
        return r

    def _get_json(self, uri, debug=False) -> Optional[dict]:
        r = self._serviceconnector.request('GET', uri, debug=debug)
        # If the resource is not found, return None ...
        if r.status_code == requests.codes.not_found:
            return None
        raise_for_status_with_detail(r)
        return r.json()

    def _request_json(self, uri, method='GET'):
        r = self._serviceconnector.request(method, uri)
        raise_for_status_with_detail(r)
        return r.json()

    @classmethod
    def from_current_cli_profile(cls: Type[T], version: str = '3', **kwargs) -> T:
        cli_cfg = get_cortex_profile()
        url, token = cli_cfg["url"], cli_cfg["token"]
        return cls(url, version, token, **kwargs)   # type: ignore # ignore until mypy properyly supports attr ...

    @classmethod
    def from_cortex_message(cls, message, version:str="3", **kwargs):
        return cls(message["apiEndpoint"], version, message["token"], **kwargs)


"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import dill
import os
import json

from pathlib import Path
from .run import Run, RemoteRun
from .properties import PropertyManager
from contextlib import closing
from datetime import datetime
from .camel import CamelResource
from typing import Dict, List
from requests.exceptions import HTTPError
from .exceptions import APIException, ConfigurationException
from .serviceconnector import _Client
from .utils import raise_for_status_with_detail


class ExperimentClient(_Client):

    """
    A client for the Cortex experiment and model management API.
    """

    URIs = {
        'experiments': 'experiments',
        'experiment': 'experiments/{experiment_name}',
        'runs': 'experiments/{experiment_name}/runs',
        'run': 'experiments/{experiment_name}/runs/{run_id}',
        'artifact': 'experiments/{experiment_name}/runs/{run_id}/artifacts/{artifact}',
        'meta': 'experiments/{experiment_name}/runs/{run_id}/meta/{meta}',
        'param': 'experiments/{experiment_name}/runs/{run_id}/params/{param}',
        'metric': 'experiments/{experiment_name}/runs/{run_id}/metrics/{metric}'
    }

    def list_experiments(self):
        r = self._serviceconnector.request(method='GET', uri=self.URIs['experiments'])
        raise_for_status_with_detail(r)
        rs = r.json()

        return rs.get('experiments', [])

    def save_experiment(self, experiment_name, **kwargs):
        body_obj = {'name': experiment_name}

        if kwargs:
            body_obj.update(kwargs)

        body = json.dumps(body_obj)
        headers = {'Content-Type': 'application/json'}
        uri = self.URIs['experiments']
        r = self._serviceconnector.request(method='POST', uri=uri, body=body, headers=headers)
        raise_for_status_with_detail(r)
        return r.json()

    def delete_experiment(self, experiment_name):
        uri = self.URIs['experiment'].format(experiment_name=experiment_name)
        r = self._serviceconnector.request(method='DELETE', uri=uri)
        raise_for_status_with_detail(r)
        rs = r.json()

        return rs.get('success', False)

    def get_experiment(self, experiment_name):
        uri = self.URIs['experiment'].format(experiment_name=experiment_name)
        r = self._serviceconnector.request(method='GET', uri=uri)
        raise_for_status_with_detail(r)

        return r.json()

    def list_runs(self, experiment_name):
        uri = self.URIs['runs'].format(experiment_name=experiment_name)
        r = self._serviceconnector.request(method='GET', uri=uri)
        raise_for_status_with_detail(r)
        rs = r.json()

        return rs.get('runs', [])

    def find_runs(self, experiment_name, filter, sort=None, limit=25):
        uri = self.URIs['runs'].format(experiment_name=experiment_name)

        # filter and limit are required query params
        params = {
            'filter': json.dumps(filter),
            'limit': limit
        }

        # Add sorting
        if sort:
            params['sort'] = json.dumps(sort)

        r = self._serviceconnector.request(method='GET', uri=uri, params=params)
        raise_for_status_with_detail(r)
        rs = r.json()

        return rs.get('runs', [])

    def delete_runs(self, experiment_name, filter=None, sort=None, limit=None):
        uri = self.URIs['runs'].format(experiment_name=experiment_name)

        params = {}

        # Add query filter
        if filter:
            params['filter'] = json.dumps(filter)

        # Add sorting
        if sort:
            params['sort'] = json.dumps(sort)

        # Add limit
        if limit:
            params['limit'] = limit

        r = self._serviceconnector.request(method='DELETE', uri=uri, params=params)
        raise_for_status_with_detail(r)
        rs = r.json()

        return rs.get('message')

    def create_run(self, experiment_name, **kwargs):
        body_obj = {}

        if kwargs:
            body_obj.update(kwargs)

        body = json.dumps(body_obj)
        headers = {'Content-Type': 'application/json'}
        uri = self.URIs['runs'].format(experiment_name=experiment_name)
        r = self._serviceconnector.request(method='POST', uri=uri, body=body, headers=headers)
        raise_for_status_with_detail(r)
        return r.json()

    def get_run(self, experiment_name, run_id):
        uri = self.URIs['run'].format(experiment_name=experiment_name, run_id=run_id)
        r = self._serviceconnector.request(method='GET', uri=uri)
        raise_for_status_with_detail(r)

        return r.json()

    def update_run(self, experiment_name, run_id, **kwargs):
        body_obj = {}

        if kwargs:
            body_obj.update(kwargs)

        body = json.dumps(body_obj)
        headers = {'Content-Type': 'application/json'}
        uri = self.URIs['run'].format(experiment_name=experiment_name, run_id=run_id)
        r = self._serviceconnector.request(method='PUT', uri=uri, body=body, headers=headers)
        raise_for_status_with_detail(r)
        rs = r.json()

        success = rs.get('success', False)
        if not success:
            raise Exception('Error updating run {}: {}'.format(run_id, rs.get('error')))
        return success

    def delete_run(self, experiment_name, run_id):
        uri = self.URIs['run'].format(experiment_name=experiment_name, run_id=run_id)
        r = self._serviceconnector.request(method='DELETE', uri=uri)
        raise_for_status_with_detail(r)
        rs = r.json()

        success = rs.get('success', False)
        if not success:
            raise Exception('Error deleting run {}: {}'.format(run_id, rs.get('error')))
        return success

    def update_meta(self, experiment_name, run_id, meta, val):
        uri = self.URIs['meta'].format(experiment_name=experiment_name, run_id=run_id, meta=meta)
        headers = {'Content-Type': 'application/json'}
        r = self._serviceconnector.request(method='PUT', uri=uri, body=json.dumps({'value': val}), headers=headers)
        raise_for_status_with_detail(r)
        rs = r.json()

        success = rs.get('success', False)
        if not success:
            raise Exception('Error updating run {} meta property {}: {}'.format(run_id, meta, rs.get('error')))
        return success

    def update_param(self, experiment_name, run_id, param, val):
        uri = self.URIs['param'].format(experiment_name=experiment_name, run_id=run_id, param=param)
        headers = {'Content-Type': 'application/json'}
        r = self._serviceconnector.request(method='PUT', uri=uri, body=json.dumps({'value': val}), headers=headers)
        raise_for_status_with_detail(r)
        rs = r.json()

        success = rs.get('success', False)
        if not success:
            raise Exception('Error updating run {} param {}: {}'.format(run_id, param, rs.get('error')))
        return success

    def update_metric(self, experiment_name, run_id, metric, val):
        uri = self.URIs['metric'].format(experiment_name=experiment_name, run_id=run_id, metric=metric)
        headers = {'Content-Type': 'application/json'}
        r = self._serviceconnector.request(method='PUT', uri=uri, body=json.dumps({'value': val}), headers=headers)
        raise_for_status_with_detail(r)
        rs = r.json()

        success = rs.get('success', False)
        if not success:
            raise Exception('Error updating run {} metric {}: {}'.format(run_id, metric, rs.get('error')))
        return success

    def update_artifact(self, experiment_name, run_id, artifact, stream):
        uri = self.URIs['artifact'].format(experiment_name=experiment_name, run_id=run_id, artifact=artifact)
        r = self._serviceconnector.request(method='PUT', uri=uri, body=stream)
        raise_for_status_with_detail(r)
        rs = r.json()

        success = rs.get('success', False)
        if not success:
            raise Exception('Error updating run {} artifact {}: {}'.format(run_id, artifact, rs.get('error')))
        return success

    def get_artifact(self, experiment_name, run_id, artifact):
        uri = self.URIs['artifact'].format(experiment_name=experiment_name, run_id=run_id, artifact=artifact)
        r = self._serviceconnector.request(method='GET', uri=uri, stream=True)
        raise_for_status_with_detail(r)

        return r.content


def _to_html(exp):
    try:
        import maya
        from jinja2 import Template
    except (ImportError, NameError):
        raise ConfigurationException(
            'The jupyter extras are required to use this, please install using "pip install cortex-python[jupyter]"')

    runs = exp.runs()

    template = """
                    <style>
                        #table1 {
                          border: solid thin;
                          border-collapse: collapse;
                        }
                        #table1 caption {
                          padding-bottom: 0.5em;
                        }
                        #table1 th,
                        #table1 td {
                          border: solid thin;
                          padding: 0.5rem 2rem;
                        }
                        #table1 td {
                          white-space: nowrap;
                        }
                        #table1 td {
                          border-style: none solid;
                          vertical-align: top;
                        }
                        #table1 th {
                          padding: 0.2em;
                          vertical-align: middle;
                          text-align: center;
                        }
                        #table1 tbody td:first-child::after {
                          content: leader(". "); '
                        }
                    </style>
                    <table id="table1">
                        <caption><b>Experiment:</b> {{name}}</caption>
                        <thead>
                        <tr>
                            <th rowspan="2">ID</th>
                            <th rowspan="2">Date</th>
                            <th rowspan="2">Took</th>
                            <th colspan="{{num_params}}" scope="colgroup">Params</th>
                            <th colspan="{{num_metrics}}" scope="colgroup">Metrics</th>
                        </tr>
                        <tr>
                            {% for param in param_names %}
                            <th>{{param}}</th>
                            {% endfor %}
                            {% for metric in metric_names %}
                            <th>{{metric}}</th>
                            {% endfor %}
                        </tr>
                        </thead>
                        <tbody>
                            {% for run in runs %}
                            <tr>
                            <td>{{run.id}}</td>
                            <td>{{maya(run.start_time)}}</td>
                            <td>{{'%.2f' % run.took}} s</td>
                            {% for param in param_names %}
                            <td>{{run.params.get(param, "&#x2011;")}}</td>
                            {% endfor %}
                            {% for metric in metric_names %}
                            <td>{{'%.6f' % run.metrics.get(metric, 0.0)}}</td>
                            {% endfor %}
                            </tr>
                            {% endfor %}
                        </tbody>
                    </table>"""

    num_params = 0
    num_metrics = 0
    param_names = set()
    metric_names = set()
    if len(runs) > 0:
        for one_run in runs:
            param_names.update(one_run.params.keys())
            num_params = len(param_names)
            metric_names.update(one_run.metrics.keys())
            num_metrics = len(metric_names)

    t = Template(template)
    return t.render(
            name=exp.name, 
            runs=runs, 
            maya=maya.MayaDT, 
            num_params=num_params, 
            param_names=sorted(list(param_names)), 
            num_metrics=num_metrics, 
            metric_names=sorted(list(metric_names))
            )


class Experiment(CamelResource):
    """
    Tracks runs, associated parameters, metrics, and artifacts of experiments.
    """

    def __init__(self, document: Dict, client: ExperimentClient):
        super().__init__(document, False)
        self._client = client
        if not self.meta:
            self.meta = {}

    @staticmethod
    def get_experiment(name, client: ExperimentClient, **kwargs):
        """
        Fetches or creates an experiment to work with.

        :param name: The name of the experiment to retrieve.
        :param client: The client instance to use.
        :return: An experiment object.
        """
        try:
            exp = client.get_experiment(name)
        except HTTPError:
            # Likely a 404, try to create a new experiment
            exp = client.save_experiment(name, **kwargs)

        return Experiment(exp, client)

    def start_run(self) -> Run:
        return RemoteRun.create(self, self._client)

    def save_run(self, run: Run):
        self._client.update_run(self.name, run.id, took=run.took, startTime=run.start_time, endTime=run.end_time)

    def reset(self, filter=None, sort=None, limit=None):
        self._client.delete_runs(self.name, filter, sort, limit)

    def set_meta(self, prop, value):
        self.meta[prop] = value
        self._client.save_experiment(self.name, **self.to_camel())

    def runs(self) -> List[Run]:
        runs = self._client.list_runs(self.name)
        return [RemoteRun.from_json(r, self) for r in runs]

    def get_run(self, run_id) -> Run:
        run = self._client.get_run(self.name, run_id)
        return RemoteRun.from_json(run, self)

    def last_run(self) -> Run:
        sort = {'endTime': -1}
        runs = self._client.find_runs(self.name, {}, sort=sort, limit=1)
        if len(runs) == 1:
            return RemoteRun.from_json(runs[0], self)
        raise APIException('Last run for experiment {} not found'.format(self.name))

    def find_runs(self, filter, sort, limit: int) -> List[Run]:
        runs = self._client.find_runs(self.name, filter or {}, sort=sort, limit=limit)
        return [RemoteRun.from_json(r, self) for r in runs]

    def load_artifact(self, run: Run, name: str):
        return dill.loads(self._client.get_artifact(self.name, run.id, name))

    def to_camel(self, camel='1.0.0'):
        return {
            'camel': camel,
            'name': self.name,
            'title': self.title,
            'description': self.description,
            'tags': self.tags or [],
            'meta': self.meta or {}
        }

    def _repr_html_(self):
        return _to_html(self)

    def display(self):
        from IPython.display import (display, HTML)
        display(HTML(self._repr_html_()))


class LocalExperiment:
    """
    Runs experiment locally, not using Cortex services.
    """
    config_file = 'config.yml'
    root_key = 'experiment'
    dir_cortex = '.cortex'
    dir_local = 'local'
    dir_artifacts = 'artifacts'
    dir_experiments = 'experiments'
    runs_key = 'runs'

    def __init__(self, name, basedir=None):
        self._name = name

        if basedir:
            self._basedir = Path(basedir)
        else:
            self._basedir = Path.home() / self.dir_cortex

        self._work_dir = self._basedir / self.dir_local / self.dir_experiments / self.name
        self._work_dir.mkdir(parents=True, exist_ok=True)
        Path(self._work_dir / self.dir_artifacts).mkdir(parents=True, exist_ok=True)

        # Initialize config
        pm = PropertyManager()
        try:
            pm.load(str(self._work_dir / self.config_file))
        except FileNotFoundError:
            pm.set('meta', {'created': str(datetime.now())})

        self._config = pm

    @property
    def name(self):
        """
        Name of the experiment.
        """
        return self._name

    def start_run(self) -> Run:
        """
        Creates a run for the experiment.
        """
        return Run(self)

    def save_run(self, run: Run):
        """
        Saves a run.

        :param run: The run you want to save.
        """
        updated_runs = []
        runs = self._config.get(self._config.join(self.root_key, self.runs_key)) or []
        replaced = False
        if len(runs) > 0:
            for r in runs:
                if r['id'] == run.id:
                    updated_runs.append(run.to_json())
                    replaced = True
                else:
                    updated_runs.append(r)

        if not replaced:
            updated_runs.append(run.to_json())

        self._config.set(self._config.join(self.root_key, self.runs_key), updated_runs)
        self._save_config()

        for name, artifact in run.artifacts.items():
            with closing(open(self.get_artifact_path(run, name), 'wb')) as f:
                dill.dump(artifact, f)

    def reset(self):
        """
        Resets the experiement, removing all associated configuration and runs.
        """
        self._config.remove_all()
        self.clean_dir(self._work_dir)
        self.clean_dir(Path(self._work_dir / self.dir_artifacts))

    def clean_dir(self, dir_to_clean):
        """
        Removes only the files from the given directory
        """
        for f in os.listdir(dir_to_clean):
             if os.path.isfile(os.path.join(dir_to_clean, f)):
                 os.remove(os.path.join(dir_to_clean, f))

    def set_meta(self, prop, value):
        """
        Associates metadata properties with the experiment.

        :param prop: The name of the metadata property to associate with the experiment.
        :param value: The value of the metadata property to associate with the experiment.
        """
        meta = self._config.get('meta')
        meta[prop] = value
        self._config.set('meta', meta)
        self._save_config()

    def runs(self) -> List[Run]:
        """
        Returns the runs associated with the experiment.
        """
        props = self._config
        runs = props.get(props.join(self.root_key, self.runs_key)) or []
        return [Run.from_json(r, self) for r in runs]

    def get_run(self, run_id: str) -> Run:
        """
        Gets a particular run from the runs in this experiment.
        """
        for r in self.runs():
            if r.id == run_id:
                return r
        return None

    def last_run(self) -> Run:
        runs = self.runs()
        if len(runs) > 0:
            return runs[-1]
        return None

    def find_runs(self, filter, sort, limit: int) -> List[Run]:
        raise NotImplementedError('find_runs is not supported in local mode')

    def _save_config(self):
        self._config.save(self._work_dir / self.config_file)

    def load_artifact(self, run: Run, name: str, extension: str = 'pk'):
        """
        Returns a particualr artifact created by the given run of the experiement.

        :param run: The run that generated the artifact requested.
        :param name: The name of the artifact.
        :param extension: An optional extension (defaults to 'pk').
        """
        artifact_file = self.get_artifact_path(run, name, extension)
        with closing(open(artifact_file, 'rb')) as f:
            return dill.load(f)

    def get_artifact_path(self, run: Run, name: str, extension: str = 'pk'):
        """
        Returns the fully qualified path to a particular artifact.

        :param run: The run that generated the artifact requested.
        :param name: The name of the artifact.
        :param extension: An optional extension (defaults to 'pk').
        """
        return self._work_dir / self.dir_artifacts / "{}_{}.{}".format(name, run.id, extension)

    def _repr_html_(self):
        return _to_html(self)

    def display(self):
        """
        Provides html output of the experiment.
        """
        try:
            from IPython.display import (display, HTML)
            display(HTML(self._repr_html_()))
        except ImportError:
            raise ConfigurationException(
                'The ipython package is required, please install it using pip install cortex-python[jupyter]')



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time


class Timer:
    """
    Used with the run class to time actions.
    """
    def __init__(self):
        self.start_time = None
        self._start_clock = None
        self.end_time = None
        self._end_clock = None
        self.interval = None

    def start(self):
        """
        Starts the timer.
        """
        self._start_clock = time.perf_counter()
        self.start_time = time.time()

    def stop(self):
        """
        Stops the timer.
        """
        self._end_clock = time.perf_counter()
        self.end_time = time.time()
        self.interval = self._end_clock - self._start_clock

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
from mocket.mockhttp import Entry
from mocket import mocketize
import pkg_resources
import requests

from cortex.serviceconnector import ServiceConnector

URL = "http://1.2.3.4:80"
VERSION = 3
TOKEN = "abc123"

def test__init__():
    s = ServiceConnector(URL, VERSION, TOKEN)

    assert s.token == TOKEN
    assert s.url == URL
    assert s.version == VERSION


def test__construct_url():
    sc = ServiceConnector(URL, VERSION, TOKEN)
    r = sc._construct_url('abc')
    expect = 'http://1.2.3.4:80/v3/{}'.format('abc')
    assert r == expect

@mocketize
def test_request():
    sc = ServiceConnector(URL, VERSION, TOKEN)
    path = 'models/events'
    url = sc._construct_url(path)
    body={"handle": 123}
    userAgentFragment = f'cortex-python/{pkg_resources.get_distribution("cortex-python").version}'
    Entry.single_register(
        Entry.POST,
        url,
        status = 200,
        body = json.dumps(body)
    )
    r = sc.request('POST', path, body)

    assert isinstance(r, requests.Response)
    assert r.status_code == 200
    assert r.json() == body
    assert userAgentFragment in r.request.headers['user-agent']


"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest
from cortex.pipeline import Pipeline


class T1:
    def transform(self, pipeline, data):
        pipeline.set_context('T1', {'name': 'T1'})
        data['foo'] = data.get('foo', 0) + 1


def T5(pipeline, data):
    pipeline.set_context('T5', {'name': 'T5'})
    data['foo'] = data.get('foo', 0) + 1


def T3(pipeline, data):
    pipeline.set_context('T3', {'name': 'T3'})
    data['bar'] = data.get('bar', 0) + 2


class PipelineTest(unittest.TestCase):

    def test_transforms(self):
        p = Pipeline('test_transforms')
        p.add_step(T1)

        class T2:
            def transform(self, pipeline, data):
                return data

        try:
            # Nested classes are not supported, expecting AttributeError
            p.add_step(T2)
            assert False # Fail
        except AttributeError:
            pass

        def T4(pipeline, data):
            return data

        # Nested function should be serialized properly
        p.add_step(T4)

        data = {}
        p.run(data)
        assert 'foo' in data
        assert 1 == data['foo']

    def test_mixed_steps(self):
        p = Pipeline('test_mixed_steps')
        p.add_step(T1)
        p.add_step(T3)

        data = {}
        p.run(data)

        assert 'foo' in data
        assert 1 == data['foo']

        assert 'bar' in data
        assert 2 == data['bar']

    def test_dumps(self):
        p = Pipeline('test_dumps')
        p.add_step(T1)
        p.add_step(T3)

        data = {}
        p.run(data)

        print(p.dumps())

    def test_load(self):
        p = Pipeline('test_load')
        p.add_step(T1)
        p.add_step(T3)

        p_obj = p.to_camel()
        p = Pipeline.load(p_obj)

        data = {}
        p.run(data)

        assert 'foo' in data
        assert 1 == data['foo']

        assert 'bar' in data
        assert 2 == data['bar']

    def test_dependencies(self):
        p1 = Pipeline('p1')
        p1.add_step(T1)

        p3 = Pipeline('p3')
        p3.add_step(T5)

        p2 = Pipeline('p2')
        p2.add_step(T3)
        p2.add_dependency(p1)
        p2.add_dependency(p3)
        p2.add_dependency(p3)   # Intentional - make sure p3 only runs once

        data = {}
        p2.run(data)

        assert 'foo' in data
        assert 2 == data['foo']

        assert 'bar' in data
        assert 2 == data['bar']

        assert p2.get_context('T1') is not None

        p2_obj = p2.to_camel()
        p2 = Pipeline.load(p2_obj)

        data = {}
        p2.run(data)

        assert 'foo' in data
        assert 2 == data['foo']

        assert 'bar' in data
        assert 2 == data['bar']

        assert p2.get_context('T1') is not None




"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import glob

from requests_toolbelt.multipart.encoder import MultipartEncoder
from requests.exceptions import HTTPError
import tenacity
import time

from .serviceconnector import _Client
from .utils import raise_for_status_with_detail


class ManagedContentClient(_Client):
    """
    A client used to access the Cortex managed content service (blob store).
    """
    URIs = {'content':  'content'}

    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Content apis only have v2..
        self._serviceconnector.version = 2

    def upload(self, key: str, stream_name: str, stream: object, content_type: str, retries: int = 1):
        """Store `stream` file in S3.

        :param key: The path where the file will be stored.
        :param stream_name: The name under which to save the `stream`.
        :param stream: The file object.
        :param content_type: the type of the file to store (e.g., `text/csv`).
        :param retries: Number of times to retry a failed request from a retryable response.
        :return: a dict with the response to request upload.

        .. NOTE: This method uses a multi-part form request; to upload very large files, use `uploadStreaming` instead.
        .. seealso: uploadStreaming()
        """
        r = tenacity.Retrying(
            stop = tenacity.stop_after_attempt(retries + 1),
            retry = tenacity.retry_if_exception(ManagedContentClient._http_request_retry_predicate)
            )
        return r.wraps(self._upload)(key, stream_name, stream, content_type)

    def _upload(self, key: str, stream_name: str, stream: object, content_type: str):
        uri = self.URIs['content']
        fields = {'key': key, 'content': (stream_name, stream, content_type)}
        data = MultipartEncoder(fields=fields)
        headers = {'Content-Type': data.content_type}
        r = self._serviceconnector.request('POST', uri, data, headers)
        raise_for_status_with_detail(r)
        return r.json()

    @staticmethod
    def _get_source_files(source):
        source_regex = os.path.join(source, '**/*')
        for path in glob.iglob(source_regex, recursive=True):
            if os.path.isfile(path):
                yield {
                    'canonical': path,
                    'relative': path[len(source):]
                }

    def upload_directory(self, source: str, destination: str, retries: int = 1):
        """Walk source directory and store in S3

        :param source: The path to the local directory.
        :param destination: Prefix to add to resulting saved directory.
        :param retries: Number of times to retry a failed request from a retryable response.
        :return: A dict with the response to request upload.
        """
        source_path = source
        if not source.endswith('/'):
            source_path = source_path + '/'

        generated_paths = ManagedContentClient._get_source_files(source_path)
        responses = []
        for path_dict in generated_paths:
            key = os.path.join(destination, path_dict.get('relative'))
            with open(path_dict.get('canonical'), 'rb') as stream:
                responses.append(self.upload_streaming(key, stream, 'application/octet-stream', retries))

        return responses

    def upload_streaming(self, key: str, stream: object, content_type: str, retries: int = 1):
        """Store `stream` file in S3.

        :param key: The path where the file will be stored.
        :param stream: The file object.
        :param content_type: The type of the file to store (e.g., `text/csv`)
        :param retries: Number of times to retry a failed request from a retryable response.
        :return: A dict with the response to request upload.
        """
        r = tenacity.Retrying(
            stop = tenacity.stop_after_attempt(retries + 1),
            retry = tenacity.retry_if_exception(ManagedContentClient._http_request_retry_predicate)
            )
        return r.wraps(self._upload_streaming)(key, stream, content_type)

    def _upload_streaming(self, key: str, stream: object, content_type: str):
        uri = self._make_content_uri(key)
        headers = {'Content-Type': content_type}
        r = self._serviceconnector.request('POST', uri, stream, headers)
        raise_for_status_with_detail(r)
        return r.json()

    def download(self, key: str, retries: int = 1):
        """Download a file from managed content (S3).

        :param key: The path of the file to retrieve.
        :param retries: Number of times to retry a failed request from a response.
        :returns: A Generator.
        """
        r = tenacity.Retrying(
            stop = tenacity.stop_after_attempt(retries + 1),
            retry = tenacity.retry_if_exception(ManagedContentClient._http_request_retry_predicate)
            )
        return r.wraps(self._download)(key)

    def _download(self, key: str):
        uri = self._make_content_uri(key)
        r = self._serviceconnector.request('GET', uri, stream=True)
        raise_for_status_with_detail(r)
        return r.raw

    def exists(self, key: str) -> bool:
        """Check that a file from managed content (S3) exists.

        :param key: The path of the file to check.
        :returns: A boolean indicating wether the file exists or not.
        """
        uri = self._make_content_uri(key)
        r = self._serviceconnector.request('HEAD', uri)
        return r.status_code == 200

    ## Private ##

    @staticmethod
    def _http_request_retry_predicate(exception: Exception) -> bool:
        if isinstance(exception, HTTPError):
            retry_after = exception.response.headers.get('Retry-After')
            if retry_after:
                ## NOTE: sleep in predicate, ugly, but how to extract this?
                time.sleep(retry_after)
                return True
        return False

    def _make_content_uri(self, key: str):
        return self.URIs['content'] + '/' + key.lstrip('/')



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .utils import get_logger
from .serviceconnector import _Client
from typing import Dict

log = get_logger(__name__)


class SessionClient(_Client):
    """
    A client for the Cortex Sessions API.
    """

    URIs = {'start': 'sessions/start', 'get': 'sessions/{session_id}', 'put': 'sessions/{session_id}', 'delete': 'sessions/{session_id}'}

    def start_session(self, ttl=None, instance_id=None, description='No description given') -> str:
        """
        Starts a new session.

        :param ttl: Resets sessions expiration; default is 15 minutes.
        :param instance_id: An optional ID that scopes this session to a deployed agent-instance
        :param description: An optional human readable description for this sessions instance
        :return: The ID of the new Session.
        """
        uri = self.URIs['start']
        params = {}
        if ttl:
            params['ttl'] = ttl
        if instance_id:
            params['instance_id'] = instance_id

        params['description'] = description

        result = self._post_json(uri, params)
        return result.get('sessionId')

    def get_session_data(self, session_id, key=None) -> Dict[str, object]:
        """
        Gets data for a specific session.

        :param session_id: The ID of the session to query.
        :param key: An optional key in the session memory; the entire session memory is returned if a key is not specified.
        :return: A dict containing the requested session data
        """
        uri = self.URIs['get'].format(session_id=session_id)
        if key:
            uri += '?key={key}'.format(key=key)

        result = self._get_json(uri) or {}
        return result.get('state', {})

    def put_session_data(self, session_id, data: Dict):
        """
        Adds data to an existing session.

        :param session_id: The ID of the session to modify.
        :param data: Dict containing the new session keys to set.

        """
        uri = self.URIs['put'].format(session_id=session_id)
        return self._post_json(uri, {'state': data})

    def delete_session(self, session_id):
        """
        Deletes a session.

        :param session_id: The ID of the session to delete.

        """
        uri = self.URIs['delete'].format(session_id=session_id)
        return self._request_json(uri, method='DELETE')


class Session:
    """
    Represents a state for a client interaction with Cortex.
    """

    def __init__(self, session_id, client: SessionClient):
        self._session_id = session_id
        self._client = client

    def get(self, key: str) -> object:
        """
        Gets the session data corresponding to the given key.

        :param key: the key to retrieve
        :return: session data corresponding to the given key
        """
        return self._client.get_session_data(self._session_id, key)

    def put(self, key: str, value: Dict):
        """
        Gets the session data corresponding to the given key.

        :param key: the key for the data to be put in the session
        :param value: the value of the data to be put in the session
        :return: a json representation of the data put in the session
        """
        return self._client.put_session_data(self._session_id, value)

    def get_all(self) -> Dict:
        """
        Gets all the data associated with the session.
        """
        return self._client.get_session_data(self._session_id)

    def delete(self):
        """
        Deletes the session from the client.
        """
        return self._client.delete_session(self._session_id)

    @staticmethod
    def start(client: SessionClient, ttl=None, instance_id=None):
        """
        Creates a new session for a given client.

        :param client: The client to associate with this session.
        :param ttl: An optional session time to live.
        :param instance_id: An optional identifier for this session; if not
        provided, the client creates a number.
        :return: A session attached to the given client.
        """
        session_id = client.start_session(ttl, instance_id)
        return Session(session_id, client)



"""
Copyright 2018 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import unittest

from mocket.mockhttp import Entry
from mocket import mocketize

from cortex.auth import AuthenticationClient

class TestAuthenticationClient(unittest.TestCase):

    def setUp(self):
        self.ac = AuthenticationClient('http://localhost:8000', 2)

    @mocketize
    def test_fetch_auth_token(self):
        # setup
        uri = self.ac.URIs['authenticate'].format('cogscale')
        url = self.ac._serviceconnector._construct_url(uri)
        body={"jwt": "123"}
        Entry.single_register(Entry.POST,
                              url,
                              status = 200,
                              body = json.dumps(body))
        # execute
        token = self.ac.fetch_auth_token('cogscale', 'marcus', 'secret')
        # test
        self.assertEqual(token, '123')




"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from setuptools import setup
from setuptools import find_packages

with open('README.md') as f:
    long_description = f.read()

setup(name='cortex-python',
      description="Python module for the CognitiveScale Cortex Cognitive Platform",
      long_description=long_description,
      long_description_content_type='text/markdown',
      version='1.4.0',
      author='CognitiveScale',
      author_email='info@cognitivescale.com',
      url='https://github.com/CognitiveScale/cortex-python',
      license='Apache License Version 2.0',
      platforms=['linux', 'osx'],
      packages=find_packages(),
      include_package_data=True,
      install_requires=[
          'requests>=2.12.4,<3',
          'requests-toolbelt==0.8.0',
          'pyjwt>=1.6.1,<2',
          'pyyaml>=5.3.1,<6',
          'cuid>=0.3,<1',
          'tenacity>=5.0.2',
          'dill>=0.2.8.2'
      ],
      extras_require={
          'viz': [
              'matplotlib>=2.2.2,<3',
              'seaborn>=0.9.0,<0.10',
              'pandas'
          ],
          'jupyter': [
              'ipython>=6.4.0,<7',
              'maya>=0.5.0',
              'jinja2'
          ],
          'builders': [
              'cortex-python-builders>=1.0.0,<2'
          ]
      },
      tests_require=[
          'mocket>=2.5.0,<3',
          'mock>=2,<3',
          'pytest>=3.2.5,<4',
          'pipdeptree'
      ],
      classifiers=[
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX',
          'Programming Language :: Python :: 3.6',
      ])



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from cortex.dataset import DatasetsClient

import json
import unittest
from mocket.mockhttp import Entry
from mocket import mocketize

from .fixtures import john_doe_token, mock_api_endpoint, build_mock_url, register_entry


class DatasetTestClient(unittest.TestCase):
    """
    Tests for the datasetclient.
    """

    def setUp(self):
        self.ds_client = DatasetsClient(mock_api_endpoint(), 3, john_doe_token())

    @mocketize
    def test_list_datasets(self):
        # mocks
        get_datasets_uri = DatasetsClient.URIs['datasets']
        get_datasets_returns = {'datasets': [{'_environmentId': 'cortex/default',
                                              '_id': '000000000000000000000000',
                                              'connectionQuery': [{'_id': '000000000000000000000000',
                                                                   'value': 'CSV',
                                                                   'name': 'contentType'},
                                                                  {'_id': '000000000000000000000000', 'value': ',', 'name': 'csv/delimiter'},
                                                                  {'_id': '000000000000000000000000',
                                                                   'value': 'true',
                                                                   'name': 'csv/headerRow'},
                                                                  {'_id': '000000000000000000000000',
                                                                   'value': '/cortex/datasets/dataset.csv/000000000000000000000000.csv',
                                                                   'name': 'key'}],
                                              'description': '',
                                              'connectionName': 'cortex/content',
                                              'parameters': [
                                                  {'format': 'int64', 'type': 'integer', 'name': 'value'},
                                                  {'type': 'string', 'name': 'value2'}
                                              ],
                                              'camel': '1.0.0',
                                              'title': 'TestDataset',
                                              'name': 'dataset.csv',
                                              'tags': [],
                                              'connectionWrite': [],
                                              '_tenantId': 'tenant',
                                              'createdAt': '2019-04-19T10:10:10.000',
                                              'updatedAt': '2019-04-19T10:10:10.000Z',
                                              '_version': 1}
                                             ]
                                }
        register_entry(Entry.GET, build_mock_url(get_datasets_uri), get_datasets_returns)
        list_datasets_expected = get_datasets_returns["datasets"]
        list_datasets_got = self.ds_client.list_datasets()
        self.assertEqual(len(list_datasets_got), 1)
        self.assertEqual(json.dumps(list_datasets_got, sort_keys=True),
                         json.dumps(list_datasets_expected, sort_keys=True))



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time
import json
from .message import Message
from .utils import get_logger
from .serviceconnector import ServiceConnector
from .camel import Document, CamelResource
from .utils import raise_for_status_with_detail

log = get_logger(__name__)


class ServiceActivation(Document):
    """
    Gets the activation status for a service activation.
    """
    URIS = {'activation': 'agents/services/activations/{activation_id}'}

    def __init__(self, activation, connector: ServiceConnector):
        super().__init__(activation, True)
        self._activation_id = activation.get('activationId')
        self._session_id = activation.get('sessionId')
        self._instance_id = activation.get('instanceId')
        self._connector = connector
        self._start = activation.get('start') or time.time()
        self._end = activation.get('end') or -1

    def result(self, timeout=0):
        now = time.time()
        end = now + timeout

        while now <= end:
            activation = self._fetch_activation()
            if activation.get('status') == 'COMPLETE':
                params = {
                    'payload': activation.get('response'),
                    'instanceId': activation.get('instanceId'),
                    'channelId': activation.get('channelId'),
                    'sessionId': activation.get('sessionId')
                }

                return Message(params)
            elif activation.get('status') == 'ERROR':
                params = {
                    'payload': activation.get('response') or activation.get('error'),
                    'instanceId': activation.get('instanceId'),
                    'channelId': activation.get('channelId'),
                    'sessionId': activation.get('sessionId'),
                    'error': True
                }

                return Message(params)
            else:
                time.sleep(1)
                now = time.time()

        raise TimeoutError('Timeout waiting for result')

    def get_status(self):
        activation = self._fetch_activation()
        return activation.get('status')

    def _fetch_activation(self):
        uri = self.URIS['activation'].format(activation_id=self._activation_id)
        r = self._connector.request('GET', uri)
        raise_for_status_with_detail(r)

        rsp = r.json()
        return rsp.get('activation')


class Agent(CamelResource):
    """
    A Cortex5 agent, which represents the deployable applications
    that aggregate the skills, inputs and outputs, and data flows.
    """
    URIS = {
        'get_agent': 'catalog/agents/{name}',
        'invoke_service': 'agents/{agent_name}/services/{service_name}'
    }

    def __init__(self, agent, connector: ServiceConnector):
        super().__init__(agent, True)
        self._connector = connector

    def invoke_service(self, service_name: str, message: Message, timeout=30):
        """
        Invokes a service, polling for a result for the number of seconds specified when the service times out.

        :param service_name: The name of the service to invoke.
        :param message: The payload for the service invocation.
        :param timeout: Number of seconds to poll for a result.
        :return: Service activation result.
        """
        activation = self.invoke_service_async(service_name, message)
        return activation.result(timeout)


    def invoke_service_async(self, service_name: str, message: Message) -> ServiceActivation:
        """
        Invokes a service.

        :param service_name: The name of the service to invoke.
        :param message: The payload for the service invocation.
        :return: Service activation.
        """
        uri = self.URIS['invoke_service'].format(agent_name=self.name, service_name=service_name)
        r = self._connector.request(method='POST', uri=uri, body=json.dumps(message.to_params()), headers={'Content-Type': 'application/json'})
        raise_for_status_with_detail(r)

        return ServiceActivation(r.json(), self._connector)

    @staticmethod
    def get_agent(name: str, connector: ServiceConnector):
        """
        Fetches an agent to work with.

        :param name: The name of the agent to retrieve.
        :param connector: A service connector.
        :return: An agent object.
        """
        uri = Agent.URIS['get_agent'].format(name=name)
        log.debug('Getting agent using URI: %s' % uri)
        r = connector.request('GET', uri)
        raise_for_status_with_detail(r)

        return Agent(r.json(), connector)



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict
from .serviceconnector import _Client


class CatalogClient(_Client):
    """
    A client for the catalog REST API.
    """

    URIs = {'types': 'catalog/types',
            'agents': 'catalog/agents',
            'skills': 'catalog/skills'
            }

    def save_schema(self, type: Dict[str, object]):
        """
        Saves a schema.

        :param type: A Cortex schema as dict.
        """
        return self._post_json(self.URIs['types'], type)

    def save_agent(self, agent: Dict[str, object]):
        """
        Saves an agent.

        :param agent: A Cortex agent as dict.
        """
        return self._post_json(self.URIs['agents'], agent)

    def save_skill(self, skill: Dict[str, object]):
        """
        Saves a skill.

        :param skill: A Cortex skill as dict.
        """
        return self._post_json(self.URIs['skills'], skill)



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest
import json

import dill
from mocket.mockhttp import Entry
from mocket import mocketize

from cortex import Cortex
from cortex.experiment import ExperimentClient
from cortex.run import RemoteRun
from .fixtures import john_doe_token, build_mock_url, mock_api_endpoint


class TestRun(unittest.TestCase):
    '''
    Test experiment checks experiment functionality
    '''
    RUN_EXP_NAME = 'unittest/exp'
    RUN_ID = '000001'

    def setUp(self):
        self.local = Cortex.local()
        self.cortex = Cortex.client(api_endpoint=mock_api_endpoint(), api_version=3, token=john_doe_token())

        # register mock for getting an expeiment from the client
        uri = ExperimentClient.URIs['experiment'].format(experiment_name=self.RUN_EXP_NAME)
        returns = {"name": self.RUN_EXP_NAME}
        Entry.single_register(Entry.GET, build_mock_url(uri), status=200, body=json.dumps(returns))

        # register mock for creating a run
        uri = ExperimentClient.URIs['runs'].format(experiment_name=self.RUN_EXP_NAME)

        returns = {"runId": self.RUN_ID}
        Entry.single_register(Entry.POST, build_mock_url(uri), status=200, body=json.dumps(returns))

    @mocketize
    def test_make_remote_run(self): 
        exp = self.cortex.experiment(self.RUN_EXP_NAME, version='3')
        r = exp.start_run()

        self.assertIsInstance(r, RemoteRun)

    @mocketize
    def test_run_get_artifact(self):
        exp = self.cortex.experiment(self.RUN_EXP_NAME, version='3')
        r = exp.start_run()

        test_artifact = {'a': 0.7071, 'b': 1.4142, 'c': 2.718}

        # get the artifact & test if it is what is expected
        ## register mock for loading an artifact
        uri = ExperimentClient.URIs['artifact'].format(
                experiment_name=self.RUN_EXP_NAME, 
                run_id=self.RUN_ID, 
                artifact='artifact'
                )
        Entry.single_register(Entry.GET, build_mock_url(uri), status=200, body=dill.dumps(test_artifact))
        result = r.get_artifact('artifact')

        self.assertEqual(test_artifact, result)
